"""External resources for SWAP model and SWAP 4.2 executables.

This module contains the paths to the SWAP model and SWAP 4.2 executables and
external resources, allowing for easy access to them throughout the package. The
following resources are available:

Modules:
    libs: The root directory of the `libs` module.
    swap_linux: The directory of the SWAP420 Linux executable.
    swap_windows: The directory of the SWAP420 Windows executable.
    crop_params: The directory of WOFOST crop parameters.

Linux and Windows executables are compiled executables distributed with the
SWAP 4.2 model through the main website of the SWAP model
(https://www.swap.alterra.nl/). The license of the source code provided with the
precompiled executables is included in the `libs` root.

The WOFOST crop parameters shipped with the package as submodule come from
GitHub repository of Allard de Wit
(https://github.com/ajwdewit/WOFOST_crop_parameters)
"""

from importlib import resources
from pathlib import Path

libs: Path = resources.files(__name__)
"""Libs module directory."""

swap: Path = libs / "swap"
"""The directory of the SWAP executable."""

swap_linux: Path = swap / "swap420"
"""The directory of the SWAP420 Linux executable."""

swap_windows: Path = swap / "swap.exe"
"""The directory of the SWAP420 Windows executable."""

crop_params: Path = libs / "WOFOST_crop_parameters"
"""The directory of WOFOST crop parameters."""

plotting: Path = libs / "plotting"
"""The directory of plotting resources."""

co2concentration: Path = libs / "co2-concentration" / "Atmospheric.co2"



"""Database integrations for pySWAP.

This module contains classes and functions that allow pySWAP to interact with
databases. Currently, there is support for the WOFOST crop database (with the
WOFOSTCropDB class) and the HDF5 file format (with the HDF5 class).

### HDF5
Thanks to the HDF5 integration, users will be able to save their models and
results to a single .h5 file. HDF5 databases are binary files that can store
large amounts of data in a structured way. This format is widely used in the
scientific community and is supported by many programming languages and
different platforms. The idea is that

### WOFOST crop database
The WOFOST crop database is a database containing crop parameters for the
WOFOST model in files in YAML format. The WOFOSTCropDB class allows users to
load a particular variaty of a crop with the corresponding parameters and update
the pySWAP models with them.

Modules:
    hdf5: Class for interacting with HDF5 files.
    cropdb: Classes that wrap the crop parameters database for
        WOFOST (A. de Wit).
"""

from pyswap.db.cropdb import CropVariety, WOFOSTCropDB
from pyswap.db.hdf5 import HDF5

__all__ = ["HDF5", "CropVariety", "SoilProfile", "SoilProfilesDB", "WOFOSTCropDB"]



# mypy: disable-error-code="call-overload, misc, override"
# Justification for exemptions:
# - call-overload and misc was raised wherever I unpacked values inside Field()
#   calls (e.g., **_UNITRANGE). This approach works correctly.
# - override was raised on model_string, because the methods do not share the
#   same signature. This was not a proirity to fix.
"""Boundary conditions settings.

Classes:

    BottomBoundary: Bottom boundary settings.
"""

from pathlib import Path as _Path
from typing import Literal as _Literal

from pydantic import (
    Field as _Field,
    PrivateAttr as _PrivateAttr,
)

from pyswap.components.tables import (
    CSEEPARR,
    DATET,
    GWLEVEL,
    HAQUIF,
    HBOT5,
    INISSOIL,
    MISC,
    QBOT2,
    QBOT4,
    QTAB,
)
from pyswap.core.basemodel import PySWAPBaseModel as _PySWAPBaseModel
from pyswap.core.fields import (
    Decimal2f as _Decimal2f,
    String as _String,
    Table as _Table,
)
from pyswap.core.valueranges import (
    UNITRANGE as _UNITRANGE,
    YEARRANGE as _YEARRANGE,
)
from pyswap.utils.mixins import (
    FileMixin as _FileMixin,
    SerializableMixin as _SerializableMixin,
    YAMLValidatorMixin as _YAMLValidatorMixin,
)

__all__ = [
    "CSEEPARR",
    "DATET",
    "GWLEVEL",
    "HAQUIF",
    "HBOT5",
    "INISSOIL",
    "MISC",
    "QBOT2",
    "QBOT4",
    "QTAB",
    "BottomBoundary",
]


class BottomBoundary(
    _PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin, _FileMixin
):
    """Bottom boundary settings.

    Technically in SWAP boundary conditions can be specified either inside the
    .swp file or in a separate .bbc file. The `swbbcfile` attribute determines
    whether the boundary conditions are written to a .bbc file.

    Attributes:
        swbbcfile (Optional[Literal[0, 1]]): Specify boundary conditions in
            current file (0) or in a separate .bbc file (1). The preferred is to
            define the boundary conditions in the .swp file. bbcfil might become
            deprecated in the future.

        swbotb (Optional[Literal[1, 2, 3, 4, 5, 6, 7, 8]]): Switch for type of
            bottom boundary.

            * 1 - prescribe groundwater level;
            * 2 - prescribe bottom flux;
            * 3 - calculate bottom flux from hydraulic head of deep aquifer;
            * 4 - calculate bottom flux as function of groundwater level;
            * 5 - prescribe soil water pressure head of bottom compartment;
            * 6 - bottom flux equals zero;
            * 7 - free drainage of soil profile;
            * 8 - free outflow at soil-air interface.

        sw2 (Optional[Literal[1, 2]]): Specify whether a sinus function or
            a table are used for the bottom flux.

            * 1 - sinus function;
            * 2 - table.

        sw3 (Optional[Literal[1, 2]]): Specify whether a sinus function or
            a table are used for the hydraulic head in the deep aquifer.

            * 1 - sinus function;
            * 2 - table.

        sw4 (Optional[Literal[0, 1]]): An extra groundwater flux can be
            specified which is added to above specified flux.

            * 0 - no extra flux;
            * 1 - extra flux.

        swbotb3resvert (Optional[Literal[0, 1]]): Switch for vertical
            hydraulic resistance between bottom boundary and groundwater level.

            * 0 - Include vertical hydraulic resistance
            * 1 - Suppress vertical hydraulic resistance

        swbotb3impl (Optional[Literal[0, 1]]): Switch for numerical solution
            of bottom flux.

            * 0 - Explicit solution (choose always when SHAPE < 1.0);
            * 1 - Implicit solution.

        swqhbot (Optional[Literal[1, 2]]): Specify whether an exponential
            relation or a table is used.

            * 1 - bottom flux is calculated with an exponential relation
            * 2 - bottom flux is derived from a table

        bbcfil (Optional[String]): Name of file with bottom boundary data
            (without .BBC extension).
        sinave (Optional[Decimal2f]): Average value of bottom flux.
        sinamp (Optional[Decimal2f]): Amplitude of bottom flux sine function.
        sinmax (Optional[Decimal2f]): Time of the year with maximum bottom flux.
        shape (Optional[Decimal2f]): Shape factor to derive average groundwater
            level.
        hdrain (Optional[Decimal2f]): Mean drain base to correct for average
            groundwater level.
        rimlay (Optional[Decimal2f]): Vertical resistance of aquitard.
        aqave (Optional[Decimal2f]): Average hydraulic head in underlaying
            aquifer.
        aqamp (Optional[Decimal2f]): Amplitude hydraulic head sinus wave.
        aqtmax (Optional[Decimal2f]): First time of the year with maximum
            hydraulic head.
        aqper (Optional[Decimal2f]): Period of hydraulic head sinus wave.
        cofqha (Optional[Decimal2f]): Coefficient A for exponential relation for
            bottom flux.
        cofqhb (Optional[Decimal2f]): Coefficient B for exponential relation for
            bottom flux.
        cofqhc (Optional[Decimal2f]): Coefficient C for exponential relation for
            bottom flux.
        gwlevel (Optional[Table]): Table with groundwater level data.
        qbot (Optional[Table]): Table with bottom flux data.
        haquif (Optional[Table]): Table with average pressure head in
            underlaying aquifer.
        qbot4 (Optional[Table]): Table with bottom flux data.
        qtab (Optional[Table]): Table with groundwater level-bottom
            flux relation.
        hbot5 (Optional[Table]): Table with the bottom compartment
            pressure head.
    """

    _extension = _PrivateAttr(default="bbc")

    swbbcfile: _Literal[0, 1] | None = None
    bbcfil: _String | None = None
    swbotb: _Literal[1, 2, 3, 4, 5, 6, 7, 8] | None = None
    sw2: _Literal[1, 2] | None = None
    sw3: _Literal[1, 2] | None = None
    sw4: _Literal[0, 1] | None = None
    swbotb3resvert: _Literal[0, 1] | None = None
    swbotb3impl: _Literal[0, 1] | None = None
    swqhbot: _Literal[1, 2] | None = None
    sinave: _Decimal2f | None = _Field(ge=-10.0, le=10.0, default=None)
    sinamp: _Decimal2f | None = _Field(ge=-10.0, le=10.0, default=None)
    sinmax: _Decimal2f | None = _Field(**_YEARRANGE, default=None)
    shape: _Decimal2f | None = _Field(**_UNITRANGE, default=None)
    hdrain: _Decimal2f | None = _Field(ge=-10000.0, le=0.0, default=None)
    rimlay: _Decimal2f | None = _Field(ge=0, le=100000.0, default=None)
    aqave: _Decimal2f | None = _Field(ge=-10000, le=1000, default=None)
    aqamp: _Decimal2f | None = _Field(ge=0, le=1000.0, default=None)
    aqtmax: _Decimal2f | None = _Field(**_YEARRANGE, default=None)
    aqper: _Decimal2f | None = _Field(**_YEARRANGE, default=None)
    cofqha: _Decimal2f | None = _Field(ge=-100.0, le=100.0, default=None)
    cofqhb: _Decimal2f | None = _Field(ge=-1.0, le=1.0, default=None)
    cofqhc: _Decimal2f | None = _Field(ge=-10.0, le=10.0, default=None)
    gwlevel: _Table | None = None
    qbot: _Table | None = None
    haquif: _Table | None = None
    qbot4: _Table | None = None
    qtab: _Table | None = None
    hbot5: _Table | None = None

    def bbc(self) -> str:
        """Return the string representing the bbc file."""
        return self._model_string(exclude={"swbbcfile", "bbcfil"})

    def _model_string(self, **kwargs) -> str:
        """Internal method to handle model string generation.

        This was implemented to avoid pydantic from raising a maximum recursion
        depth error when calling the model_string method from the super class.
        """
        return super().model_string(**kwargs)

    def model_string(self, **kwargs) -> str:
        """Override model_string method to handle the swbbcfile attribute.

        This method is called in the final serialization step, when each section
        is converted into a string representation. So, depending on the
        swbbcfile attribute, this function will return:

            - a full section string representation, as for when all boundary
                conditions are included in the .swp file, or;
            - it will only include swbbcfile and bbcfil (the name of the file
                when the other parameters are defined). In that case, the other
                parameters are written to a separate .bbc file using the
                write_bbc method.
        """
        if self.swbbcfile == 1:
            return super().model_string(include={"swbbcfile", "bbcfil"}, **kwargs)
        else:
            return super().model_string()

    def write_bbc(self, path: _Path):
        """Write bottom boundary conditions to a .bbc file.

        This method is only available when the swbbcfile attribute is set to 1.
        Writes entire setion settings (except swbbcfile and bbcfil, defined in
        the .swp file) to a separate .bbc file.

        Parameters:
            path (Path): Path to the directory where the .bbc file will be
                saved.
        """
        if self.swbbcfile != 1:
            msg = "Bottom boundary conditions are not set to be written to a .bbc file."
            raise ValueError(msg)

        self.save_file(string=self.bbc(), fname=self.bbcfil, path=path)



# mypy: disable-error-code="attr-defined"
# attr-defined is disabled because it was easier to implement part of the
# functionality of one mixin in the base class. Could be considered to be
# fixed later.
"""Base models inherited by all pySWAP models.

A lot of functionality can be abstracted away in the base models. This way, the
code is more DRY and easier to maintain. The base models are used to enforce
the correct data types and structure of the input data. They also provide
methods to convert the data to the format required by the SWAP model.

Classes defined here are based on Pydantic BaseModel and Pandera DataFrameModel.
Both are meant to ensure the correct data types and structure of the input data,
as successful validation means smooth execution of the SWAP model. Particularily
important when run as a submitted job on an HPC.

Classes:
    BaseModel: Base class for pySWAP models. Inherits from Pydantic BaseModel.
    BaseTableModel: Base class for pySWAP models that validate pandas
        DataFrames. Inherits from Pandera DataFrameModel.
"""

from __future__ import annotations

from typing import Any

import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame
from pydantic import BaseModel, ConfigDict, field_validator

from pyswap.core.defaults import ADDITIONAL_SWITCHES


class PySWAPBaseModel(BaseModel):
    """Base class for pySWAP models.

    Methods:
        __setattr__: Overriden method to silently ignore assignment of frozen
            fields.
        update: Update the model with new values from a dictionary.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="ignore",
        populate_by_name=True,
    )

    def __setattr__(self, name, value):
        """Silently ignore assignment of frozen fields.

        This method is overridden to silently ignore assignment of frozen fields
        to avoid errors when an old swp files is read.
        """
        _class = type(self)

        if name in _class.model_fields and _class.model_fields[name].frozen:
            return
        super().__setattr__(name, value)

    def update(self, new: dict, inplace: bool = False, no_validate: bool = False):
        """Update the model with new values.

        Given dictionary of values is first filtered to include only the fields
        that exist in the model. The model is then updated with the new values.
        The updated model is returned (either new or updated self).

        Parameters:
            new (dict): Dictionary with new values.
            inplace (bool): If True, update the model in place.
        """

        updated_model = self.model_validate(dict(self) | new)

        if not inplace:
            # added this for the case when the user loads a model from the
            # classic ASCII files. Then the .update() method is used, but not
            # all the attributes will be available immediatelly. Full validation
            # will still be performed upon model run.
            if no_validate:
                updated_model._validation = False
            else:
                updated_model._validation = True
            updated_model.validate_with_yaml() if hasattr(
                updated_model, "validate_with_yaml"
            ) else None
            return updated_model.model_copy(deep=True)

        else:
            for field, value in updated_model:
                setattr(self, field, value)
            if no_validate:
                updated_model._validation = False
            else:
                updated_model._validation = True
            self.validate_with_yaml() if hasattr(
                updated_model, "validate_with_yaml"
            ) else None

            return self

    @field_validator("*", mode="before")
    @classmethod
    def convert_switches(cls, value: Any, info: Any) -> Any:
        """Convert switch values to integers.

        This method was necessary to ensure that loading models from ASCII files
        would work. It could be improved to include literals that do not start
        with "sw" as well.

        !!! note:
            This should be eventually replaced by a custom Switch field type handling
            serialization and deserialization.
        """
        if (
            (info.field_name.startswith("sw") or info.field_name in ADDITIONAL_SWITCHES)
            and info.field_name != "swap_ver"
            and value
        ):
            try:
                return int(value)
            except ValueError:
                return value
        return value


class BaseTableModel(pa.DataFrameModel):
    """Base model for pandas DataFrames.

    Methods:
        create: Create a validated DataFrame from a dictionary.
    """

    class Config:
        coerce = True

    @classmethod
    def create(cls, data: dict, columns: list | None = None) -> DataFrame:
        df = pd.DataFrame(data=data)
        if columns:
            df.columns = columns
        else:
            df.columns = df.columns.str.upper()
        validated_df = cls.validate(df)
        return validated_df

    @classmethod
    def update(cls, table, new: dict):
        # Update the table with new values
        table_upd = table.to_dict("list")
        table_upd.update(new)
        return cls.create(table_upd)



import pandas as pd

import pyswap.components.crop as crp


def test_model_serialization(simple_serializable_model):
    tabular_data = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]})

    model = simple_serializable_model(
        table=tabular_data,
        arrays=tabular_data,
        daymonth="12 01",
        stringlist=["a", "b", "c"],
        floatlist=[1.0, 2.0, 3.0],
        intlist=[1, 2, 3],
        datelist=[pd.Timestamp("2021-01-01"), pd.Timestamp("2021-01-02")],
        string="string",
        decimal2f=1.0,
        decimal3f=1.0,
        decimal4f=1.0,
    )

    expected_output = " col1  col2\n    1     4\n    2     5\n    3     6\n\nARRAYS = \n1 4\n2 5\n3 6\n\nDAYMONTH = 12 01\nSTRINGLIST = 'a,b,c'\nFLOATLIST = 1.00 2.00 3.00\nINTLIST = 1 2 3\nDATELIST = \n2021-01-01\n2021-01-02\nSTRING = 'string'\nDECIMAL2F = 1.00\nDECIMAL3F = 1.000\nDECIMAL4F = 1.0000"

    assert (
        model.model_string() == expected_output
    ), f"Expected: \n {expected_output} \n but got \n {model.model_string()}"


def test_table_update():
    table = crp.CROPROTATION.create(
        {
            "CROPSTART": ["2000-04-01"],
            "CROPEND": ["2000-10-31"],
            "CROPFIL": ["'maizes'"],
            "CROPTYPE": [1],
        }
    )

    # Test if method runs without errors
    crp.CROPROTATION.update(table, {"CROPTYPE": [2]})


if __name__ == "__main__":
    test_table_update()



# %%
from datetime import date as dt

import pyswap as psp
from pyswap import testcase
from pyswap.db import WOFOSTCropDB

# %%


def _make_hupselbrook():
    # %% Basic settings of the model

    ml = psp.Model()

    meta = psp.components.Metadata(
        author="John Doe",
        institution="University of Somewhere",
        email="john.doe@somewhere.com",
        project="psp test - hupselbrook",
        swap_ver="4.2",
    )

    ml.metadata = meta

    simset = psp.components.simsettings.GeneralSettings(
        tstart="2002-01-01",
        tend="2004-12-31",
        extensions=["vap", "blc", "sba", "inc", "csv"],
        nprintday=1,
        swerror=1,
        swscre=0,
        swmonth=1,
        swyrvar=0,
        datefix="31 12",
        inlist_csv=[
            "rain",
            "irrig",
            "interc",
            "runoff",
            "drainage",
            "dstor",
            "epot",
            "eact",
            "tpot",
            "tact",
            "qbottom",
            "gwl",
        ],
    )

    ml.generalsettings = simset

    # %% Meteorology section

    meteo_location = psp.gis.Location(lat=52.0, lon=21.0, alt=10.0)

    meteo_data = psp.components.meteorology.MetFile(
        metfil="283.met", content=testcase.load_met("hupselbrook")
    )

    meteo = psp.components.meteorology.Meteorology(
        meteo_location=meteo_location,
        swetr=0,
        metfile=meteo_data,
        swdivide=1,
        swmetdetail=0,
        altw=10.0,
        angstroma=0.25,
        angstromb=0.5,
    )

    ml.meteorology = meteo

    # %% Creating the .crp file for maize (fixed crop)

    maize_prep = psp.components.crop.Preparation(
        swprep=0, swsow=0, swgerm=0, dvsend=3.0, swharv=0
    )

    scheduled_irrigation = psp.components.irrigation.ScheduledIrrigation(schedule=0)

    DVS = [0.0, 0.3, 0.5, 0.7, 1.0, 1.4, 2.0]

    maize_gctb = psp.components.crop.GCTB.create(
        {
            "DVS": DVS,
            "LAI": [0.05, 0.14, 0.61, 4.10, 5.00, 5.80, 5.20],
        }
    )

    maize_chtb = psp.components.crop.CHTB.create(
        {
            "DVS": DVS,
            "CH": [1.0, 15.0, 40.0, 140.0, 170.0, 180.0, 175.0],
        }
    )

    maize_rdtb = psp.components.crop.RDTB.create(
        {
            "DVS": [0.0, 0.3, 0.5, 0.7, 1.0, 2.0],
            "RD": [5.0, 20.0, 50.0, 80.0, 90.0, 100.0],
        }
    )

    maize_rdctb = psp.components.crop.RDCTB.create(
        {
            "RRD": [0.0, 1.0],
            "RDENS": [1.0, 0.0],
        }
    )

    maize_cropdev_settings = psp.components.crop.CropDevelopmentSettingsFixed(
        idev=1,
        lcc=168,
        kdif=0.6,
        kdir=0.75,
        swgc=1,
        gctb=maize_gctb,
        swcf=2,
        chtb=maize_chtb,
        albedo=0.23,
        rsc=61.0,
        rsw=0.0,
        swrd=1,
        rdtb=maize_rdtb,
        rdctb=maize_rdctb,
    )

    maize_ox_stress = psp.components.crop.OxygenStress(
        swoxygen=1,
        swwrtnonox=0,
        aeratecrit=0.5,
        hlim1=-15.0,
        hlim2u=-30.0,
        hlim2l=-30.0,
    )

    maize_dr_stress = psp.components.crop.DroughtStress(
        swdrought=1,
        hlim3h=-325.0,
        hlim3l=-600.0,
        hlim4=-8000.0,
        adcrh=0.5,
        adcrl=0.1,
    )

    # serves both, Fixed crop and WOFOST
    maize_interception = psp.components.crop.Interception(swinter=1, cofab=0.25)

    crpmaize = psp.components.crop.CropFile(
        name="maizes",
        prep=maize_prep,
        scheduledirrigation=scheduled_irrigation,
        cropdev_settings=maize_cropdev_settings,
        oxygenstress=maize_ox_stress,
        droughtstress=maize_dr_stress,
        interception=maize_interception,
    )

    # %% Creating .crp file for potato (WOFOST)

    potato_prep = psp.components.crop.Preparation(
        swprep=0,
        swsow=0,
        swgerm=2,
        tsumemeopt=170.0,
        tbasem=3.0,
        teffmx=18.0,
        hdrygerm=-500.0,
        hwetgerm=-100.0,
        zgerm=-10.0,
        agerm=203.0,
        dvsend=2.0,
        swharv=0,
    )

    potato_chtb = psp.components.crop.CHTB.create(
        {
            "DVS": [0.0, 1.0, 2.0],
            "CH": [
                1.0,
                40.0,
                50.0,
            ],
        }
    )

    potato_rdctb = psp.components.crop.RDCTB.create(
        {
            "RRD": [0.0, 1.0],
            "RDENS": [1.0, 0.0],
        }
    )

    # Load the crop database
    db = WOFOSTCropDB()
    potato = db.load_crop_file("potato")
    potato_params = potato.get_variety("Potato_701")

    potato_cropdev_settings = psp.components.crop.CropDevelopmentSettingsWOFOST(
        wofost_variety=potato_params,
        swcf=2,
        chtb=potato_chtb,
        idsl=0,
        albedo=0.19,
        laiem=0.0589,
        ssa=0.0,
        kdif=1.0,
        rsc=207.0,
        rsw=0.0,
        kdir=0.75,
        eff=0.45,
        swrd=2,
        rdc=50.0,
        swdmi2rd=1,
        rdctb=potato_rdctb,
    )

    potato_cropdev_settings.update_from_wofost()

    potato_ox_stress = psp.components.crop.OxygenStress(
        swoxygen=1,
        swwrtnonox=1,
        aeratecrit=0.5,
        hlim1=-10.0,
        hlim2u=-25.0,
        hlim2l=-25.0,
        swrootradius=2,
        root_radiuso2=0.00015,
    )

    potato_dr_stress = psp.components.crop.DroughtStress(
        swdrought=1,
        hlim3h=-300.0,
        hlim3l=-500.0,
        hlim4=-10000.0,
        adcrh=0.5,
        adcrl=0.1,
    )

    crppotato = psp.components.crop.CropFile(
        name="potatod",
        prep=potato_prep,
        cropdev_settings=potato_cropdev_settings,
        oxygenstress=potato_ox_stress,
        droughtstress=potato_dr_stress,
        # shared with the fixed crop settings
        interception=maize_interception,
        scheduledirrigation=scheduled_irrigation,
    )

    # %% Grass crp file
    grass_chtb = psp.components.crop.CHTB.create(
        {
            "DNR": [0.0, 180.0, 366.0],
            "CH": [12.0, 12.0, 12.0],
        }
    )

    grass_slatb = psp.components.crop.SLATB.create(
        {
            "DNR": [1.00, 80.00, 300.00, 366.00],
            "SLA": [0.0015, 0.0015, 0.0020, 0.0020],
        }
    )

    grass_amaxtb = psp.components.crop.AMAXTB.create(
        {
            "DNR": [1.00, 95.00, 200.00, 275.00, 366.00],
            "AMAX": [40.00, 40.00, 35.00, 25.00, 25.00],
        }
    )

    grass_tmpftb = psp.components.crop.TMPFTB.create(
        {
            "TAVD": [0.00, 5.00, 15.00, 25.00, 40.00],
            "TMPF": [0.00, 0.70, 1.00, 1.00, 0.00],
        }
    )
    grass_tmnftb = psp.components.crop.TMNFTB.create(
        {
            "TMNR": [0.0, 4.0],
            "TMNF": [0.0, 1.0],
        }
    )

    grass_rfsetb = psp.components.crop.RFSETB.create(
        {
            "DNR": [1.00, 366.00],
            "RFSE": [1.0000, 1.0000],
        }
    )

    grass_frtb = psp.components.crop.FRTB.create(
        {
            "DNR": [1.00, 366.00],
            "FR": [0.3000, 0.3000],
        }
    )

    grass_fltb = psp.components.crop.FLTB.create(
        {
            "DNR": [1.00, 366.00],
            "FL": [0.6000, 0.6000],
        }
    )

    grass_fstb = psp.components.crop.FSTB.create(
        {
            "DNR": [1.00, 366.00],
            "FS": [0.4000, 0.4000],
        }
    )

    grass_rdrrtb = psp.components.crop.RDRRTB.create(
        {
            "DNR": [1.0, 180.0, 366.0],
            "RDRR": [0.0, 0.02, 0.02],
        }
    )

    grass_rdrstb = psp.components.crop.RDRSTB.create(
        {
            "DNR": [1.0, 180.0, 366.0],
            "RDRS": [0.0, 0.02, 0.02],
        }
    )

    grass_rlwtb = psp.components.crop.RLWTB.create(
        {
            "RW": [300.00, 2500.00],
            "RL": [20.0, 40.0],
        }
    )

    grass_rdctb = psp.components.crop.RDCTB.create(
        {
            "RRD": [0.0, 1.0],
            "RDENS": [1.0, 0.0],
        }
    )

    grass_settings = psp.components.crop.CropDevelopmentSettingsGrass(
        swcf=2,
        chtb=grass_chtb,
        albedo=0.23,
        rsc=100.0,
        rsw=0.0,
        tdwi=1000.00,
        laiem=0.63000,
        rgrlai=0.00700,
        swtsum=1,
        ssa=0.0004,
        span=30.00,
        tbase=0.00,
        slatb=grass_slatb,
        kdif=0.60,
        kdir=0.75,
        eff=0.50,
        amaxtb=grass_amaxtb,
        tmpftb=grass_tmpftb,
        tmnftb=grass_tmnftb,
        cvl=0.6850,
        cvr=0.6940,
        cvs=0.6620,
        q10=2.0000,
        rml=0.0300,
        rmr=0.0150,
        rms=0.0150,
        rfsetb=grass_rfsetb,
        frtb=grass_frtb,
        fltb=grass_fltb,
        fstb=grass_fstb,
        perdl=0.050,
        rdrrtb=grass_rdrrtb,
        rdrstb=grass_rdrstb,
        swrd=3,
        swdmi2rd=1,
        rlwtb=grass_rlwtb,
        wrtmax=3000.0,
        swrdc=0,
        rdctb=grass_rdctb,
    )

    grass_ox_stress = psp.components.crop.OxygenStress(
        swoxygen=1, hlim1=0.0, hlim2u=1.0, hlim2l=-1.0, swwrtnonox=0
    )

    grass_drought_stress = psp.components.crop.DroughtStress(
        swdrought=1,
        swjarvis=4,
        alphcrit=0.7,
        hlim3h=-200.0,
        hlim3l=-800.0,
        hlim4=-8000.0,
        adcrh=0.5,
        adcrl=0.1,
    )

    grass_salt_stress = psp.components.crop.SaltStress(swsalinity=0)

    grass_interception = psp.components.crop.Interception(swinter=1, cofab=0.25)

    grass_co2 = psp.components.crop.CO2Correction(swco2=0)

    grass_dmmowtb = psp.components.crop.DMMOWTB.create(
        {
            "DNR": [120.0, 152.0, 182.0, 213.0, 366.0],
            "DMMOW": [4700.0, 3700.0, 3200.0, 2700.0, 2700.0],
        }
    )

    grass_dmmowdelay = psp.components.crop.DMMOWDELAY.create(
        {
            "DMMOWDELAY": [0.0, 2000.0, 4000.0],
            "DAYDELAY": [2, 3, 4],
        }
    )

    grass_management = psp.components.crop.GrasslandManagement(
        seqgrazmow=[2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2],
        swharvest=1,
        swdmmow=2,
        dmmowtb=grass_dmmowtb,
        maxdaymow=42,
        swlossmow=0,
        mowrest=700.0,
        dmmowdelay=grass_dmmowdelay,
        swpotrelmf=1,
        relmf=0.90,
    )

    grass_irrigation = psp.components.irrigation.ScheduledIrrigation(schedule=0)

    crpgrass = psp.components.crop.CropFile(
        name="grassd",
        cropdev_settings=grass_settings,
        oxygenstress=grass_ox_stress,
        droughtstress=grass_drought_stress,
        saltstress=grass_salt_stress,
        interception=grass_interception,
        co2correction=grass_co2,
        grasslandmanagement=grass_management,
        scheduledirrigation=grass_irrigation,
    )

    # %% Creating the main Crop object

    croprotation = psp.components.crop.CROPROTATION.create(
        {
            "CROPSTART": [dt(2002, 5, 1), dt(2003, 5, 10), dt(2004, 1, 1)],
            "CROPEND": [dt(2002, 10, 15), dt(2003, 9, 29), dt(2004, 12, 31)],
            "CROPFIL": ["'maizes'", "'potatod'", "'grassd'"],
            "CROPTYPE": [1, 2, 3],
        }
    )

    crop = psp.components.crop.Crop(
        swcrop=1,
        rds=200.0,
        croprotation=croprotation,
        cropfiles={"maizes": crpmaize, "potatod": crppotato, "grassd": crpgrass},
    )

    ml.crop = crop

    # %% irrigation setup

    irrig_events = psp.components.irrigation.IRRIGEVENTS.create(
        {
            "IRDATE": ["2002-01-05"],
            "IRDEPTH": [5.0],
            "IRCONC": [1000.0],
            "IRTYPE": [1],
        }
    )

    fixed_irrigation = psp.components.irrigation.FixedIrrigation(
        swirfix=1, swirgfil=0, irrigevents=irrig_events
    )

    ml.fixedirrigation = fixed_irrigation

    # %% Soil moisture setup

    soilmoisture = psp.components.soilwater.SoilMoisture(swinco=2, gwli=-75.0)
    ml.soilmoisture = soilmoisture

    # %% surface flow settings

    surfaceflow = psp.components.soilwater.SurfaceFlow(
        swpondmx=0, pondmx=0.2, rsro=0.5, rsroexp=1.0, swrunon=0
    )

    ml.surfaceflow = surfaceflow

    # %% evaporation settings

    evaporation = psp.components.soilwater.Evaporation(
        cfevappond=1.25, swcfbs=0, rsoil=30.0, swredu=1, cofredbl=0.35, rsigni=0.5
    )

    ml.evaporation = evaporation

    # %% setting soil profile

    soil_profile = psp.components.soilwater.SOILPROFILE.create(
        {
            "ISUBLAY": [1, 2, 3, 4],
            "ISOILLAY": [1, 1, 2, 2],
            "HSUBLAY": [10.0, 20.0, 30.0, 140.0],
            "HCOMP": [1.0, 5.0, 5.0, 10.0],
            "NCOMP": [10, 4, 6, 14],
        }
    )

    soil_hydraulic_functions = psp.components.soilwater.SOILHYDRFUNC.create(
        {
            "ORES": [0.01, 0.02],
            "OSAT": [0.42, 0.38],
            "ALFA": [0.0276, 0.0213],
            "NPAR": [1.491, 1.951],
            "KSATFIT": [12.52, 12.68],
            "LEXP": [-1.060, 0.168],
            "ALFAW": [0.0542, 0.0426],
            "H_ENPR": [0.0, 0.0],
            "KSATEXM": [12.52, 12.68],
            "BDENS": [1315.0, 1315.0],
        }
    )

    soilprofile = psp.components.soilwater.SoilProfile(
        swsophy=0,
        soilprofile=soil_profile,
        swhyst=0,
        tau=0.2,
        soilhydrfunc=soil_hydraulic_functions,
        swmacro=0,
    )

    ml.soilprofile = soilprofile

    # %% drainage settings

    dra = psp.components.drainage.DraFile(
        dramet=2,
        swdivd=1,
        cofani=[1.0, 1.0],
        swdislay=0,
        lm2=11.0,
        shape=0.8,
        wetper=30.0,
        zbotdr=-80.0,
        entres=20.0,
        ipos=2,
        basegw=-200.0,
        khtop=25.0,
    )
    drainage = psp.components.drainage.Drainage(swdra=1, drafile=dra)

    ml.lateraldrainage = drainage

    # %% bottom boundary

    bottom_boundary = psp.components.boundary.BottomBoundary(swbbcfile=0, swbotb=6)

    ml.bottomboundary = bottom_boundary

    return ml



"""Functions to fine tune the serializatino of pySWAP objects to SWAP
formatted ASCII.

More complex serialization logic which would be unwieldy to implement directly
in the Annotated field definitions (pyswap.core.fields module) as lambda
functions are defined in the serializers module (pyswap.core.serializers). These
are functions that convert objects to strings in the valid SWAP format.

Serializers in this module:
    serialize_table: Convert a DataFrame to a string.
    serialize_arrays: Convert a DataFrame to a string without headers and
        newline in front.
    serialize_csv_table: Convert a DataFrame to a string in CSV format.
    serialize_object_list: Convert a list of objects to a string.
    serialize_day_month: Convert a date object to a string with just the day
        and month.

"""

from datetime import date

from pandas import DataFrame


def serialize_table(table: DataFrame) -> str:
    """Convert the DataFrame to a string.

    Arguments:
        table: The DataFrame to be serialized.

    Result:
        >>> ' A  B\n 1  4\n 2  5\n 3  6\n'
    """
    return f"{table.to_string(index=False)}\n"


def serialize_arrays(table: DataFrame) -> str:
    """Convert the DataFrame to a string without headers and newline in front.

    Arguments:
        table: The DataFrame to be serialized.

    Result:
        >>> 'ARRAYS = \n1 4\n2 5\n3 6\n\n'
    """
    return f"\n{table.to_string(index=False, header=False)}\n"


def serialize_day_month(value: date) -> str:
    """Serialize a date object to a string with just the day and month.

    Arguments:
        value: The date object to be serialized.

    Result:
        >>> '01 01'
    """
    return value.strftime("%d %m")


def serialize_decimal(precision: int):
    def decimal_serializer(v, info):
        return f"{round(v, precision):.{precision}f}"

    return decimal_serializer



"""Interact with ASCII files.

Functions:
    open_ascii: Open an ASCII file and detect its encoding.
    save_ascii: Save a string to an ASCII file.
"""

from pathlib import Path

import chardet


def open_ascii(file_path: Path) -> str:
    """Open file and detect encoding.

    Arguments:
        file_path (str): Path to the file to be opened.
    """
    with open(file_path, "rb") as f:
        raw_data = f.read()
    encoding = chardet.detect(raw_data)["encoding"]

    return raw_data.decode(encoding)


def save_ascii(
    string: str,
    fname: str,
    path: str,
    mode: str = "w",
    extension: str | None = None,
    encoding: str = "ascii",
) -> None:
    """
    Saves a string to a file with a given extension.

    Parameters:
        string (str): The string to be saved to a file.
        extension (str): The extension that the file should have (e.g. 'txt', 'csv', etc.).
        fname (str): The name of the file.
        path (str): The path where the file should be saved.
        mode (str): The mode in which the file should be opened (e.g. 'w' for write, 'a' for append, etc.).
        encoding (str): The encoding to use for the file (default is 'ascii').

    Returns:
        None
    """

    if extension is not None:
        fname = f"{fname}.{extension}"

    with open(f"{path}/{fname}", f"{mode}", encoding=f"{encoding}") as f:
        f.write(string)



# ruff: noqa: S607
# Those rules are not really useful in this package.

"""The cli module is supposed to help in structuring the direcotries of created models and enforce
best practices in documenting. It creates a modular structure (with __init__.py files) what can be helpful when
writing scripts. This way, modules from the scripts can be directly imported into the main.py or main.ipynb"""

import json
import shutil
import subprocess
from pathlib import Path

import typer

app = typer.Typer()


def dict_to_custom_string(data):
    lines = [f"    {key}='{value}'," for key, value in data.items()]
    return "\n".join(lines)


def make_script(models_dir, basic_code_to_write_path, basic_code_to_write) -> str:
    main_script_path = models_dir / "main.py"

    try:
        if not main_script_path.exists():
            main_script_path.touch()

        main_script_path.write_text(
            basic_code_to_write_path.read_text().format(
                formatted_string=basic_code_to_write
            )
        )

        return print("Succesfully created main.py.")
    except Exception as e:
        return print(f"Error creating folder or file: {e}")


def make_notebook(models_dir, basic_code_to_write, templates_path, attrs) -> None:
    notebook_template_path = templates_path / "notebook.json"
    notebook_path = models_dir / "main.ipynb"

    with open(notebook_template_path, encoding="utf-8") as f:
        notebook_content = json.load(f)
    try:
        notebook_content["cells"][0]["source"][0] = f"# {attrs['project']}\n"
        notebook_content["cells"][2]["source"][2] = basic_code_to_write

        with open(notebook_path, "w", encoding="utf-8") as f:
            json.dump(notebook_content, f, indent=4)

        print(f"Notebook {notebook_path} created successfully.")
    except Exception as e:
        print(f"Error creating Jupyter Notebook: {e}")

    return None


def copy_readme(templates_path, project_root, use_pixi=True, attrs=None):
    """Copy and customize README template based on whether pixi is used."""
    if use_pixi:
        template_file = templates_path / "README_template"
        pixi_instructions_file = templates_path / "pixi_instructions.md"

        readme_content = template_file.read_text()
        pixi_instructions = pixi_instructions_file.read_text()

        pixi_structure = "\n├── pixi.toml               # Pixi dependency management"

        readme_content = readme_content.format(
            project=attrs.get("project", "Project"),
            pixi_structure=pixi_structure,
            pixi_instructions=pixi_instructions
        )

        readme_path = project_root / "README.md"
        readme_path.write_text(readme_content)
    else:
        template_file = templates_path / "README"
        readme_content = template_file.read_text()

        readme_content = readme_content.format(
            project=attrs.get("project", "Project"),
            pixi_structure="",
            pixi_instructions=""
        )

        readme_path = project_root / "README"
        readme_path.write_text(readme_content)

    return "Successfully created README in the root directory."


def create_pixi_toml(templates_path, project_root, attrs):
    """Create pixi.toml file from template."""
    template_file = templates_path / "pixi.toml"
    pixi_content = template_file.read_text()

    # Format the template with user attributes
    pixi_content = pixi_content.format(
        author=attrs.get("author", "Unknown"),
        email=attrs.get("email", "unknown@example.com"),
        project=attrs.get("project", "pyswap-project")
    )

    pixi_path = project_root / "pixi.toml"
    pixi_path.write_text(pixi_content)

    return "Successfully created pixi.toml file."


def copy_crop_parameter_yaml(templates_path, models_dir):
    template_file = templates_path / "crop_tables_template.yaml"
    shutil.copy(template_file, models_dir)
    return "Successfully created crop_tables_template.yaml in the models directory."


def create_inits(project_root, models_dir, scripts_dir):
    (project_root / "__init__.py").touch()
    (models_dir / "__init__.py").touch()
    (scripts_dir / "__init__.py").touch()


def init_git_repo(project_root, use_pixi=False):
    try:
        result = subprocess.run(
            ["git", "init", str(project_root)],
            shell=False,
            check=True,
            capture_output=True,
            text=True,
        )
        print(result.stdout.strip())

        # Create a .gitignore file
        gitignore_path = project_root / ".gitignore"
        with open(gitignore_path, "w") as f:
            f.write("# Python\n")
            f.write("__pycache__/\n")
            f.write("*.pyc\n")
            f.write("*.pyo\n")
            f.write("*.pyd\n")
            f.write(".env\n")
            f.write("venv/\n")
            f.write("env/\n")
            f.write("*.env\n")
            f.write("*.ipynb_checkpoints/\n")
            f.write("data/\n")

            if use_pixi:
                f.write("\n# Pixi\n")
                f.write(".pixi/\n")
                f.write("pixi.lock\n")

        print(f"Created .gitignore file at {gitignore_path}")
    except Exception as e:
        print(f"Error initializing Git repository or creating .gitignore: {e}")


@app.command()
def init(
    script: bool = False,
    notebook: bool = True,
    pixi: bool = typer.Option(True, "--pixi/--no-pixi", help="Include Pixi configuration for dependency management (default: True)")
):
    """Prompt the user to enter their information and create a User class."""
    attrs = {
        "project": typer.prompt("Project name"),
        "swap_ver": typer.prompt("SWAP version used"),
        "author": typer.prompt("Author first/last name"),
        "institution": typer.prompt("Your last institution"),
        "email": typer.prompt("Your email address"),
        "comment": typer.prompt("Any comments?", default=""),
    }

    folder_name = typer.prompt("Choose a folder name", default=attrs.get("project"))

    # Defining paths and creating folders.
    templates_path = Path(__file__).resolve().parent / "templates"
    project_root = Path.cwd() / folder_name

    basic_code_to_write_path = templates_path / "script.txt"
    basic_code_to_write = dict_to_custom_string(attrs)

    folders_to_create = ["models", "scripts", "data", "tests"]
    folders_to_create_paths = [project_root / folder for folder in folders_to_create]

    [folder.mkdir(parents=True, exist_ok=True) for folder in folders_to_create_paths]

    # Dealing with files.
    copy_readme(templates_path, project_root, use_pixi=pixi, attrs=attrs)
    create_inits(
        project_root=project_root,
        models_dir=folders_to_create_paths[0],
        scripts_dir=folders_to_create_paths[1],
    )

    # Create pixi.toml if requested
    if pixi:
        create_pixi_toml(templates_path, project_root, attrs)
        print("Created pixi.toml for dependency management.")

    if script:
        make_script(
            folders_to_create_paths[0], basic_code_to_write_path, basic_code_to_write
        )

    if notebook:
        make_notebook(
            folders_to_create_paths[0], basic_code_to_write, templates_path, attrs
        )

    init_git_repo(project_root, use_pixi=pixi)


@app.command()
def modify(script: bool = False, notebook: bool = True):
    print("Executing modify routine...")


@app.command()
def get_swap(
    version: str = typer.Option(
        "4.2.0", "--version", "-v", help="SWAP version to download"
    ),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force re-download even if executable exists"
    ),
    verbose: bool = typer.Option(
        True, "--verbose/--quiet", help="Enable verbose output"
    ),
):
    """Download and install SWAP executable."""
    from pyswap.utils.executables import get_swap as _get_swap

    try:
        exe_path = _get_swap(version=version, force=force, verbose=verbose)
        if verbose:
            typer.echo(f"Success! SWAP executable ready at: {exe_path}")
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1) from e


@app.command()
def upload_swap(
    file_path: str = typer.Argument(help="Path to the SWAP executable file to install"),
    version: str = typer.Argument(
        help="Version identifier for the uploaded executable"
    ),
    force: bool = typer.Option(
        False, "--force", "-f", help="Force replace existing executable"
    ),
    verbose: bool = typer.Option(
        True, "--verbose/--quiet", help="Enable verbose output"
    ),
):
    """Install SWAP executable from a local file."""
    from pyswap.utils.executables import upload_swap as _upload_swap

    try:
        exe_path = _upload_swap(
            file_path=file_path, version=version, force=force, verbose=verbose
        )
        if verbose:
            typer.echo(f"Success! SWAP executable installed at: {exe_path}")
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1) from e


@app.command()
def check_swap(
    verbose: bool = typer.Option(
        True, "--verbose/--quiet", help="Enable verbose output"
    ),
):
    """Check if SWAP executable is available and working."""
    from pyswap.utils.executables import check_swap as _check_swap

    if _check_swap(verbose=verbose):
        typer.echo("✓ SWAP is ready to use!")
    else:
        typer.echo(
            "✗ SWAP is not available. Run 'pyswap get-swap' to install.", err=True
        )
        raise typer.Exit(1)


@app.command()
def remove_swap(
    verbose: bool = typer.Option(
        True, "--verbose/--quiet", help="Enable verbose output"
    ),
):
    """Remove SWAP executable from package directory."""
    from pyswap.utils.executables import remove_swap as _remove_swap

    if _remove_swap(verbose=verbose):
        typer.echo("SWAP executable removed successfully")
    else:
        typer.echo("Failed to remove SWAP executable", err=True)
        raise typer.Exit(1)


@app.command()
def info():
    """Display information about pySWAP and SWAP setup."""
    from pyswap.utils.executables import show_info

    show_info()

if __name__ == "__main__":
    app()



from typing import Literal

from .grassgrowth import _make_grassgrowth
from .hupselbrook import _make_hupselbrook
from .macroporeflow import _make_macroporeflow
from .oxygenstress import _make_oxygenstress
from .salinitystress import _make_salinitystress
from .simple_test_model import _make_simple_test_model
from .surfacewater import _make_surfacewater


def get(
    case: Literal[
        "hupselbrook",
        "oxygenstress",
        "grassgrowth",
        "macroporeflow",
        "salinitystress",
        "surfacewater",
        "simple_test_model",
    ],
):
    cases = {
        "hupselbrook": _make_hupselbrook,
        "oxygenstress": _make_oxygenstress,
        "grassgrowth": _make_grassgrowth,
        "macroporeflow": _make_macroporeflow,
        "salinitystress": _make_salinitystress,
        "surfacewater": _make_surfacewater,
        "simple_test_model": _make_simple_test_model,
    }

    return cases[case]()



# ruff: noqa: SIM210
# mypy: disable-error-code="operator"
# The operator error was being raised on the method that was matching the
# dictionarties in _parse_ascii_file. This was not a priority to fix.

import re
from pathlib import Path
from typing import Literal as _Literal

from pyswap import Model
from pyswap.components.boundary import BottomBoundary
from pyswap.components.crop import (
    CO2Correction,
    CompensateRWUStress,
    Crop,
    CropDevelopmentSettingsFixed,
    CropDevelopmentSettingsGrass,
    CropDevelopmentSettingsWOFOST,
    CropFile,
    DroughtStress,
    GrasslandManagement,
    Interception,
    OxygenStress,
    Preparation,
    SaltStress,
)
from pyswap.components.drainage import DraFile, Drainage, Flux
from pyswap.components.irrigation import FixedIrrigation, ScheduledIrrigation
from pyswap.components.meteorology import Meteorology
from pyswap.components.simsettings import GeneralSettings, RichardsSettings
from pyswap.components.soilwater import (
    Evaporation,
    SnowAndFrost,
    SoilMoisture,
    SoilProfile,
    SurfaceFlow,
)
from pyswap.components.transport import HeatFlow, SoluteTransport
from pyswap.core.basemodel import PySWAPBaseModel
from pyswap.core.defaults import EXTENSION_SWITCHES
from pyswap.core.io.io_ascii import open_ascii
from pyswap.core.io.process_ascii import parse_ascii_file

__all__ = ["load_swp"]


def load_swp(path: Path, metadata: PySWAPBaseModel) -> Model:
    """Load a SWAP model from a .swp file.

    Parameters:
        path (Path): Path to the .swp file.

    Returns:
        PySWAPBaseModel: The loaded model.
    """
    # finish this one up. Essentialle you need to pop the extension switches
    # from the main dictionary and then create the extension list. The names
    # have to be handled properly.
    # params = _parse_ascii_file(path)
    swp = open_ascii(path)
    params = parse_ascii_file(swp)
    # from among the parameters parsed from the ascii file, pop the switches
    extension_switches = {
        key: params.pop(key) for key in EXTENSION_SWITCHES if key in params
    }
    # create the extension list with only those switches that are = 1. get rid of the "sw" prefix

    extension_list = [
        key[2:]
        for key, value in extension_switches.items()
        if isinstance(value, int | str) and int(value) == 1
    ]

    # model definition
    model_setup = {
        "generalsettings": GeneralSettings(extensions=extension_list),
        "meteorology": Meteorology(),
        "crop": Crop(),
        "fixedirrigation": FixedIrrigation(),
        "soilmoisture": SoilMoisture(),
        "surfaceflow": SurfaceFlow(),
        "evaporation": Evaporation(),
        "soilprofile": SoilProfile(),
        "snowandfrost": SnowAndFrost(),
        "richards": RichardsSettings(),
        "lateraldrainage": Drainage(),
        "bottomboundary": BottomBoundary(),
        "heatflow": HeatFlow(),
        "solutetransport": SoluteTransport(),
    }

    for value in model_setup.values():
        value.update(params, inplace=True)

    ml = Model(metadata=metadata, **model_setup)

    return ml


def load_dra(path: Path):
    dra = open_ascii(path)
    params = parse_ascii_file(dra)

    flux_objects_startwith = [
        "drares",
        "infres",
        "swallo",
        "l",
        "zbotdr",
        "swdtyp",
        "datowltb",
    ]

    flux_objects = {
        k: v
        for k, v in params.items()
        if any(re.match(f"{prefix}[1-5]$", k) for prefix in flux_objects_startwith)
    }
    other_params = {k: v for k, v in params.items() if k not in flux_objects}

    flux = Flux(**flux_objects)
    dra = DraFile(**other_params, fluxes=flux)

    return dra


def load_crp(path: Path, crptype: _Literal["fixed", "wofost", "grass"], name: str):
    crp = open_ascii(path)
    params = parse_ascii_file(crp, grass=True if crptype == "grass" else False)

    cropfile_setup = {
        "name": name,
        "prep": Preparation(),
        "cropdev_settings": CropDevelopmentSettingsFixed()
        if crptype == "fixed"
        else CropDevelopmentSettingsWOFOST()
        if crptype == "wofost"
        else CropDevelopmentSettingsGrass(),
        "oxygenstress": OxygenStress(),
        "droughtstress": DroughtStress(),
        "saltstress": SaltStress(),
        "compensaterwu": CompensateRWUStress(),
        "interception": Interception(),
        "scheduledirrigation": ScheduledIrrigation(),
        "grasslandmanagement": GrasslandManagement(),
        "co2correction": CO2Correction(),
    }

    for value in cropfile_setup.values():
        if isinstance(value, str):
            continue
        if isinstance(value, PySWAPBaseModel):
            value.update(new=params, inplace=True)
        else:
            continue

    crp = CropFile(**cropfile_setup)

    return crp


def load_bbc(path: Path, bottomboundary: BottomBoundary | None = None):
    """Load the bottom boundary conditions from a .bbc file.

    Bottom boundary conditions are stored in the same class. Therefore this
    function can either return a new instance of the class or update an existing
    one.
    """
    bbc = open_ascii(path)
    params = parse_ascii_file(bbc)

    if bottomboundary is None:
        bottomboundary = BottomBoundary()

    botbound = bottomboundary.update(params)

    return botbound



# mypy: disable-error-code="call-overload, misc, type-arg"
# The type-arg here causes problems with passing the actual type to the alias
# Subsection. This however helps to type hinting, so it has to stay that way.
"""Crop settings and crop files for SWAP model.

Similar to the .dra or .swp files, the .crp file is a configuration file for the SWAP model.
The classes in this module represent distincs sections of the .crp file. The main class is the
`CropFile` class which holds the settings for the crop simulation.

SWAP has three modes for crop simulations which users define in the CROPROTATION _table in the .swp file:

    * 1 - simple crop settings - use CropDevelopmentSettingsFixed
    * 2 - detailed, WOFOST general settings - use CropDevelopmentSettingsWOFOST
    * 3 - dynamic grass growth model - use CropDevelopmentSettingsGrass

For each choice, the .crp file will look different. Therefore, multiple classes
 are defined in this module to deal with those different settings.

Classes:
    CropFile: Class for the .crp file.
    CropDevelopmentSettingsWOFOST: Class for the crop development settings in WOFOST.
    CropDevelopmentSettingsFixed: Class for the fixed crop development settings.
    CropDevelopmentSettingsGrass: Class for the grass crop development settings.
    OxygenStress: Class for the oxygen stress settings.
    DroughtStress: Class for the drought stress settings.
    SaltStress: Class for the salt stress settings.
    CompensateRWUStress: Class for the compensate root water uptake stress settings.
    Interception: Class for the interception settings.
    CO2Correction: Class for the CO2 correction settings.
    ScheduledIrrigation: Class for the scheduled irrigation settings.
    Preparation: Class for the preparation settings.
"""

from typing import (
    Any,
    Literal as _Literal,
)

from pydantic import (
    Field as _Field,
    PrivateAttr as _PrivateAttr,
)

from pyswap.components.irrigation import ScheduledIrrigation as _ScheduledIrrigation
from pyswap.components.tables import (
    AMAXTB,
    CFTB,
    CHTB,
    CROPROTATION,
    DMGRZTB,
    DMMOWDELAY,
    DMMOWTB,
    DTSMTB,
    FLTB,
    FOTB,
    FRTB,
    FSTB,
    GCTB,
    KYTB,
    LSDA,
    LSDATB,
    LSDBTB,
    MRFTB,
    RDCTB,
    RDRRTB,
    RDRSTB,
    RDTB,
    RFSETB,
    RLWTB,
    SLATB,
    TMNFTB,
    TMPFTB,
    VERNRTB,
    WRTB,
)
from pyswap.core.basemodel import PySWAPBaseModel as _PySWAPBaseModel
from pyswap.core.fields import (
    Arrays as _Arrays,
    Decimal2f as _Decimal2f,
    IntList as _IntList,
    Subsection as _Subsection,
    Table as _Table,
)
from pyswap.core.valueranges import (
    UNITRANGE as _UNITRANGE,
    YEARRANGE as _YEARRANGE,
)
from pyswap.db.cropdb import CropVariety as _CropVariety
from pyswap.utils.mixins import (
    FileMixin as _FileMixin,
    SerializableMixin as _SerializableMixin,
    YAMLUpdateMixin as _YAMLUpdateMixin,
    YAMLValidatorMixin as _YAMLValidatorMixin,
)

__all__ = [
    "AMAXTB",
    "CFTB",
    "CHTB",
    "CROPROTATION",
    "DMGRZTB",
    "DMMOWDELAY",
    "DMMOWTB",
    "DTSMTB",
    "FLTB",
    "FOTB",
    "FRTB",
    "FSTB",
    "GCTB",
    "KYTB",
    "LSDA",
    "LSDATB",
    "LSDBTB",
    "MRFTB",
    "RDCTB",
    "RDRRTB",
    "RDRSTB",
    "RDTB",
    "RFSETB",
    "RLWTB",
    "SLATB",
    "TMNFTB",
    "TMPFTB",
    "VERNRTB",
    "WRTB",
    "CO2Correction",
    "CompensateRWUStress",
    "Crop",
    "CropDevelopmentSettingsFixed",
    "CropDevelopmentSettingsGrass",
    "CropDevelopmentSettingsWOFOST",
    "CropFile",
    "DroughtStress",
    "Interception",
    "OxygenStress",
    "Preparation",
    "SaltStress",
    "_CropDevelopmentSettings",
]


class _CropDevelopmentSettings(
    _PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin, _YAMLUpdateMixin
):
    """Crop development settings.

    !!! note:

        CropDevelopmentSettings is a base class for the different crop
        development settings (type 1, 2 and 3). Crop parameters can be read from
        WOFOST database. Because some names of the same parameters are different
        between the wofost and swap templates, the alias parameter is used to
        rename the parameters in serialization to SWAP compatible .crp file.
        Currently it applies to TSUM1 and TSUM2.

    Attributes:
        wofost_variety (CropVariety): Crop variety settings.
        swcf (Literal[1, 2]): Choose between crop factor and crop height for simulation of
            * 1 - Crop factor, when using ETref from meteo or Penman-Monteith.
            * 2 - Crop height, when using Penman-Monteith with actual crop height, albedo and canopy resistance.
        cftb (Optional[_Table]): Table with crop factors [0..2 -] as a function of development stage.
        chtb (Optional[_Table]): Table with crop height [0..1e4 cm] as a function of development stage.
        albedo (Optional[float]): Crop reflection coefficient [0..1.0 -].
        rsc (Optional[float]): Minimum canopy resistance [0..1e6 s/m].
        rsw (Optional[float]): Canopy resistance of intercepted water [0..1e6 s/m].
        tsumea (float): Temperature sum from emergence to anthesis [0..1e4 degrees C].
        tsumam (float): Temperature sum from anthesis to maturity [1..1e4 degrees C].
        tbase (Optional[float]): Start value of temperature sum [-10..30 degrees C].
        kdif (float): Extinction coefficient for diffuse visible light [0..2 -].
        kdir (float): Extinction coefficient for direct visible light [0..2 -].
        swrd (Optional[Literal[1, 2, 3]]): Switch development of root growth.
            * 1 - Root growth depends on development stage.
            * 2 - Root growth depends on maximum daily increase.
            * 3 - Root growth depends on available root biomass.
        rdtb (Optional [_Arrays]): Rooting depth [0..1000 cm] as a function of development stage [0..2 -].
        rdi (float): Initial rooting depth [0..1000 cm].
        rri (float): Maximum daily increase in rooting depth [0..100 cm].
        rdc (float): Maximum rooting depth of particular crop [0..1000 cm].
        swdmi2rd (Optional[Literal[0, 1]]): Switch for calculation rooting depth.
            * 0 - Rooting depth increase is related to availability assimilates for roots.
            * 1 - Rooting depth increase is related to relative dry matter increase.
        rlwtb (Optional _Arrays]): rooting depth [0..5000 cm] as function of root weight [0..5000 kg DM/ha].
        wrtmax (float): Maximum root weight [0..1e5 kg DM/ha].
        swrdc (Literal[0, 1]): Switch for calculation of relative root density.
            * 0 - Root density is not modified.
            * 1 - Root density is modified based on root water extraction.
        TODO: add parameters related to swdrc=1: fgwrt, fdwrt, wrtmin
        rdctb (_Arrays): root density [0..1 -] as function of relative rooting depth [0..1 -].
    """

    # add in model config that additional attributes are allowed
    # model_config = _ConfigDict(
    #     extra="allow"
    # )

    wofost_variety: Any | None = _Field(default=None, exclude=True)

    swcf: _Literal[1, 2] | None = None
    cftb: _Table | None = None
    chtb: _Table | None = None
    albedo: _Decimal2f | None = _Field(default=None, **_UNITRANGE)
    rsc: _Decimal2f | None = _Field(default=None, ge=0.0, le=1.0e6)
    rsw: _Decimal2f | None = _Field(default=None, ge=0.0, le=1.0e6)
    tsum1: _Decimal2f | None = _Field(alias="tsumea", default=None, ge=0.0, le=1.0e4)
    tsum2: _Decimal2f | None = _Field(alias="tsumam", default=None, ge=0.0, le=1.0e4)
    tbase: _Decimal2f | None = _Field(default=None, ge=-10.0, le=30.0)
    kdif: _Decimal2f | None = _Field(default=None, ge=0.0, le=2.0)
    kdir: _Decimal2f | None = _Field(default=None, ge=0.0, le=2.0)
    swrd: _Literal[1, 2, 3] | None = None
    rdtb: _Arrays | None = None
    rdi: _Decimal2f | None = _Field(default=None, ge=0.0, le=1000.0)
    rri: _Decimal2f | None = _Field(default=None, ge=0.0, le=100.0)
    rdc: _Decimal2f | None = _Field(default=None, ge=0.0, le=1000.0)
    swdmi2rd: _Literal[0, 1] | None = None
    rlwtb: _Arrays | None = None
    wrtmax: _Decimal2f | None = _Field(default=None, ge=0.0, le=1.0e5)
    swrdc: _Literal[0, 1] | None = None
    rdctb: _Arrays | None = None


class CropDevelopmentSettingsWOFOST(_CropDevelopmentSettings):
    """Additional settings as defined for the WOFOST model.

    Attributes:
        wofost_variety (CropVariety): Crop variety settings.
        swcf (Literal[1, 2]): Choose between crop factor and crop height for simulation of
            * 1 - Crop factor, when using ETref from meteo or Penman-Monteith.
            * 2 - Crop height, when using Penman-Monteith with actual crop height, albedo and canopy resistance.
        cftb (Optional[_Table]): Table with crop factors [0..2 -] as a function of development stage.
        chtb (Optional[_Table]): Table with crop height [0..1e4 cm] as a function of development stage.
        albedo (Optional[float]): Crop reflection coefficient [0..1.0 -].
        rsc (Optional[float]): Minimum canopy resistance [0..1e6 s/m].
        rsw (Optional[float]): Canopy resistance of intercepted water [0..1e6 s/m].
        tsumea (float): Temperature sum from emergence to anthesis [0..1e4 degrees C].
        tsumam (float): Temperature sum from anthesis to maturity [1..1e4 degrees C].
        tbase (Optional[float]): Start value of temperature sum [-10..30 degrees C].
        kdif (float): Extinction coefficient for diffuse visible light [0..2 -].
        kdir (float): Extinction coefficient for direct visible light [0..2 -].
        swrd (Optional[Literal[1, 2, 3]]): Switch development of root growth.
            * 1 - Root growth depends on development stage.
            * 2 - Root growth depends on maximum daily increase.
            * 3 - Root growth depends on available root biomass.
        rdtb (Optional [_Arrays]): Rooting depth [0..1000 cm] as a function of development stage [0..2 -].
        rdi (float): Initial rooting depth [0..1000 cm].
        rri (float): Maximum daily increase in rooting depth [0..100 cm].
        rdc (float): Maximum rooting depth of particular crop [0..1000 cm].
        swdmi2rd (Optional[Literal[0, 1]]): Switch for calculation rooting depth.
            * 0 - Rooting depth increase is related to availability assimilates for roots.
            * 1 - Rooting depth increase is related to relative dry matter increase.
        rlwtb (Optional _Arrays]): rooting depth [0..5000 cm] as function of root weight [0..5000 kg DM/ha].
        wrtmax (float): Maximum root weight [0..1e5 kg DM/ha].
        swrdc (Literal[0, 1]): Switch for calculation of relative root density.
            * 0 - Root density is not modified.
            * 1 - Root density is modified based on root water extraction.
        TODO: add parameters related to swdrc=1: fgwrt, fdwrt, wrtmin
        rdctb (_Arrays): root density [0..1 -] as function of relative rooting depth [0..1 -].
        idsl (Literal[0, 1, 2]): Switch for crop development.
            * 0 - Depends on temperature
            * 1 - Depends on temperature and daylength
            * 2 - Depends on temperature, daylength and vernalisation factor
        dtsmtb (_Arrays): List increase in temperature sum [0..60 degrees C] as function of daily average temperature.
        dlo (Optional[float]): Optimum day length for crop development [0..24 hr].
        dlc (Optional[float]): Minimum day length [0..24 hr].
        vernsat (Optional[float]): Saturated vernalisation requirement.
        vernbase (Optional[float]): Base vernalisation requirement.
        verndvs (Optional[float]): Critical development stage after which the effect of vernalisation is halted.
        vernrtb (Optional [_Arrays]): _Table with rate of vernalisation as function of average air temperature. In WOFOST it's called this way. Aliased
            to verntb for SWAP.
        tdwi (float): Initial total crop dry weight [0..10000 kg/ha].
        laiem (float): Leaf area index at emergence [0..10 m2/m2].
        rgrlai (float): Maximum relative increase in LAI [0..1 m2/m2/d].
        spa (float): Specific pod area [0..1 ha/kg].
        ssa (float): Specific stem area [0..1 ha/kg].
        span (float): Life span under leaves under optimum conditions [0..366 d].
        slatb (_Arrays): List specific leaf area [0..1 ha/kg] as function of crop development stage.
        eff (float): Light use efficiency for real leaf [0..10 kg/ha/hr/(J m2 s)].
        amaxtb (_Arrays): List maximum CO2 assimilation rate [0..100 kg/ha/hr] as function of development stage.
        tmpftb (_Arrays): List reduction factor of AMAX [-] as function of average day temperature.
        tmnftb (_Arrays): List reduction factor of AMAX [-] as function of minimum day temperature.
        cvo (float): Efficiency of conversion into storage organs [0..1 kg/kg].
        cvl (float): Efficiency of conversion into leaves [0..1 kg/kg].
        cvr (float): Efficiency of conversion into roots [0..1 kg/kg].
        cvs (float): Efficiency of conversion into stems [0..1 kg/kg].
        q10 (float): Increase in respiration rate with temperature [0..5 /10 degrees C].
        rml (float): Maintenance respiration rate of leaves [0..1 kgCH2O/kg/d].
        rmo (float): Maintenance respiration rate of storage organs [0..1 kgCH2O/kg/d].
        rmr (float): Maintenance respiration rate of roots [0..1 kgCH2O/kg/d].
        rms (float): Maintenance respiration rate of stems [0..1 kgCH2O/kg/d].
        rfsetb (_Arrays): List reduction factor of senescence [0..2 -] as function of development stage.
        frtb (_Arrays): List fraction of total dry matter increase partitioned to the roots [0..1 kg/kg] as function of development stage.
        fltb (_Arrays): List fraction of total above ground dry matter increase partitioned to the leaves [0..1 kg/kg] as function of development stage.
        fstb (_Arrays): List fraction of total above ground dry matter increase partitioned to the stems [0..1 kg/kg] as function of development stage.
        fotb (_Arrays): List fraction of total above ground dry matter increase partitioned to the storage organs [0..1 kg/kg] as function of development stage.
        perdl (float): Maximum relative death rate of leaves due to water stress [0..3 /d].
        rdrrtb (_Arrays): List relative death rates of roots [0..1 kg/kg/d] as function of development stage.
        rdrstb (_Arrays): List relative death rates of stems [0..1 kg/kg/d] as function of development stage.
    """

    idsl: _Literal[0, 1, 2] | None = None
    dtsmtb: _Arrays | None = None
    dlo: float | None = _Field(default=None, ge=0.0, le=24.0)
    dlc: float | None = _Field(default=None, ge=0.0, le=24.0)
    vernsat: float | None = _Field(default=None, ge=0.0, le=100.0)
    vernbase: float | None = _Field(default=None, ge=0.0, le=100.0)
    verndvs: float | None = _Field(default=None, ge=0.0, le=0.3)
    vernrtb: _Arrays | None = _Field(default=None, alias="verntb")
    tdwi: float | None = _Field(default=None, ge=0.0, le=10_000)
    laiem: float | None = _Field(default=None, ge=0.0, le=10)
    rgrlai: float | None = _Field(default=None, **_UNITRANGE)
    spa: float | None = _Field(**_UNITRANGE, default=None)
    ssa: float | None = _Field(default=None, **_UNITRANGE)
    span: float | None = _Field(default=None, **_YEARRANGE)
    slatb: _Arrays | None = None
    eff: float | None = _Field(default=None, ge=0.0, le=10.0)
    amaxtb: _Arrays | None = None
    tmpftb: _Arrays | None = None
    tmnftb: _Arrays | None = None
    cvo: float | None = _Field(default=None, **_UNITRANGE)
    cvl: float | None = _Field(default=None, **_UNITRANGE)
    cvr: float | None = _Field(default=None, **_UNITRANGE)
    cvs: float | None = _Field(default=None, **_UNITRANGE)
    q10: float | None = _Field(default=None, ge=0.0, le=5.0)
    rml: float | None = _Field(default=None, **_UNITRANGE)
    rmo: float | None | None = _Field(**_UNITRANGE, default=None)
    rmr: float | None = _Field(default=None, **_UNITRANGE)
    rms: float | None = _Field(default=None, **_UNITRANGE)
    rfsetb: _Arrays | None = None
    frtb: _Arrays | None = None
    fltb: _Arrays | None = None
    fstb: _Arrays | None = None
    fotb: _Arrays | None = None
    perdl: float | None = _Field(default=None, ge=0.0, le=3.0)
    rdrrtb: _Arrays | None = None
    rdrstb: _Arrays | None = None


class CropDevelopmentSettingsFixed(_CropDevelopmentSettings):
    """Fixed crop development settings (Additionaly to CropDevelopmentSettings).

    Attributes:
        swcf (Literal[1, 2]): Choose between crop factor and crop height for simulation of
            * 1 - Crop factor, when using ETref from meteo or Penman-Monteith.
            * 2 - Crop height, when using Penman-Monteith with actual crop height, albedo and canopy resistance.
        cftb (Optional[_Table]): Table with crop factors [0..2 -] as a function of development stage.
        chtb (Optional[_Table]): Table with crop height [0..1e4 cm] as a function of development stage.
        albedo (Optional[float]): Crop reflection coefficient [0..1.0 -].
        rsc (Optional[float]): Minimum canopy resistance [0..1e6 s/m].
        rsw (Optional[float]): Canopy resistance of intercepted water [0..1e6 s/m].
        tsumea (float): Temperature sum from emergence to anthesis [0..1e4 degrees C].
        tsumam (float): Temperature sum from anthesis to maturity [1..1e4 degrees C].
        tbase (Optional[float]): Start value of temperature sum [-10..30 degrees C].
        kdif (float): Extinction coefficient for diffuse visible light [0..2 -].
        kdir (float): Extinction coefficient for direct visible light [0..2 -].
        swrd (Optional[Literal[1, 2, 3]]): Switch development of root growth.
            * 1 - Root growth depends on development stage.
            * 2 - Root growth depends on maximum daily increase.
            * 3 - Root growth depends on available root biomass.
        rdtb (Optional [_Arrays]): Rooting depth [0..1000 cm] as a function of development stage [0..2 -].
        rdi (float): Initial rooting depth [0..1000 cm].
        rri (float): Maximum daily increase in rooting depth [0..100 cm].
        rdc (float): Maximum rooting depth of particular crop [0..1000 cm].
        swdmi2rd (Optional[Literal[0, 1]]): Switch for calculation rooting depth.
            * 0 - Rooting depth increase is related to availability assimilates for roots.
            * 1 - Rooting depth increase is related to relative dry matter increase.
        rlwtb (Optional _Arrays]): rooting depth [0..5000 cm] as function of root weight [0..5000 kg DM/ha].
        wrtmax (float): Maximum root weight [0..1e5 kg DM/ha].
        swrdc (Literal[0, 1]): Switch for calculation of relative root density.
            * 0 - Root density is not modified.
            * 1 - Root density is modified based on root water extraction.
        TODO: add parameters related to swdrc=1: fgwrt, fdwrt, wrtmin
        rdctb (_Arrays): root density [0..1 -] as function of relative rooting depth [0..1 -].
        idev (Literal[1, 2]): Duration of crop growing period

            * 1 - Duration is fixed
            * 2 - Duration is variable

        lcc (Optional[int]): Duration of the crop growing period
        swgc (Literal[1, 2]): Choose between Leaf Area Index or Soil Cover Fraction

            * 1 - LAI
            * 2 - SCF

        gctb  _Arrays): Soil Cover Fraction as a function of development stage
    """

    idev: _Literal[1, 2] | None = None
    lcc: int | None = _Field(default=None, **_YEARRANGE)
    swgc: _Literal[1, 2] | None = None
    gctb: _Arrays | None = None
    kytb: _Arrays | None = None


class CropDevelopmentSettingsGrass(CropDevelopmentSettingsWOFOST):
    """Crop development settings specific to grass growth.

    Attributes:
        swtsum (Literal[0, 1, 2]): Select either sum air temperatures or soil temperature at particular depth

            * 0 - no delay of start grass growth
            * 1 - start of grass growth based on sum air temperatures > 200 degree C
            * 2 - start of grass growth based on soil temperature at particular depth

        tsumtemp (Optional[float]): Specific stem area [0..1 ha/kg, R]
        tsumdepth (Optional[float]): Life span under leaves under optimum conditions [0..366 d, R]
        tsumtime (Optional[float]): Lower threshold temperature for ageing of leaves [-10..30 degree C, R]
    """

    swtsum: _Literal[0, 1, 2] | None = None
    tsumtemp: float | None = None
    tsumdepth: float | None = None
    tsumtime: float | None = None


class OxygenStress(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Oxygen stress settings for .crp file.

    Attributes:
        swoxygen (Literal[0, 1, 2]): Switch for oxygen stress

            * 0 - No oxygen stress
            * 1 - Oxygen stress according to Feddes et al. (1978)
            * 2 - Oxygen stress according to Bartholomeus et al. (2008)

        swoxygentype (Optional[Literal[1, 2]]): switch for physical processes or repro. functions to calculate oxygen stress

            * 1 - physical processes
            * 2 - reproduction functions

        swwrtnonox (Literal[0, 1]): Switch for checking aerobic conditions in root zone to stop root(zone) development
        aeratecrit (Optional[float]): Threshold to stop root extension in case of oxygenstress; 0.0 maximum oxygen stress
        hlim1 (Optional[float]): No water extraction at higher pressure heads
        hlim2u (Optional[float]): H below which optimum water extr. starts for top layer
        hlim2l (Optional[float]): H below which optimum water extr. starts for sub layer
        q10_microbial (Optional[float]): Relative increase in microbial respiration at temperature increase of 10 C
        specific_resp_humus (Optional[float]): Respiration rate of humus at 25 C
        srl (Optional[float]): Specific root length
        swrootradius (Optional[Literal[1, 2]]): Switch for calculation of root radius

            * 1 - Calculate root radius
            * 2 - Root radius given in an input file

        dry_mat_cont_roots (Optional[float]): Dry matter content of roots
        air_filled_root_por (Optional[float]): Air filled root porosity
        spec_weight_root_tissue (Optional[float]): Specific weight of non-airfilled root tissue
        var_a (Optional[float]): Variance of root radius
        root_radiuso2 (Optional[float]): Root radius for oxygen stress module
        q10_root (Optional[float]): Relative increase in root respiration at temperature increase of 10 oC
        f_senes (Optional[float]): Reduction factor for senescence, used for maintenance respiration
        c_mroot (Optional[float]): Maintenance coefficient of root
        _table_max_resp_factor (Optional[_Table]): Ratio root total respiration / maintenance respiration as a function of development stage
        _table_dvs_w_root_ss (Optional[_Table]): List dry weight of roots at soil surface as a function of development stage

    TODO: Find a way to validate the parameters that are required when the
    croptype=1 and swoxygen=2 (currently I cannot access the croptype parameter)
    >> move it to the Model class validation at the end, when all the params are available
    """

    swoxygen: _Literal[0, 1, 2] | None = None
    swwrtnonox: _Literal[0, 1] | None = None
    swoxygentype: _Literal[1, 2] | None = None
    aeratecrit: float | None = _Field(default=None, ge=0.0001, le=1.0)
    hlim1: float | None = _Field(default=None, ge=-100.0, le=100.0)
    hlim2u: float | None = _Field(default=None, ge=-1000.0, le=100.0)
    hlim2l: float | None = _Field(default=None, ge=-1000.0, le=100.0)
    q10_microbial: float | None = _Field(default=None, ge=1.0, le=4.0)
    specific_resp_humus: float | None = _Field(default=None, **_UNITRANGE)
    srl: float | None = _Field(default=None, ge=0.0, le=1.0e10)
    swrootradius: _Literal[1, 2] | None = None
    dry_mat_cont_roots: float | None = _Field(default=None, **_UNITRANGE)
    air_filled_root_por: float | None = _Field(default=None, **_UNITRANGE)
    spec_weight_root_tissue: float | None = _Field(default=None, ge=0.0, le=1.0e5)
    var_a: float | None = _Field(default=None, **_UNITRANGE)
    root_radiuso2: float | None = _Field(default=None, ge=1.0e-6, le=0.1)
    q10_root: float | None = _Field(default=None, ge=1.0, le=4.0)
    f_senes: float | None = _Field(default=None, **_UNITRANGE)
    c_mroot: float | None = _Field(default=None, **_UNITRANGE)
    mrftb: _Arrays | None = None
    wrtb: _Arrays | None = None


class DroughtStress(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Drought stress settings for .crp file.

    Attributes:
        swdrought (Literal[1, 2]): Switch for drought stress

            * 1 - Drought stress according to Feddes et al. (1978)
            * 2 - rought stress according to De Jong van Lier et al. (2008)

        swjarvis (Optional[Literal[0, 1, 2, 3, 4]]): _DEPRECATED_ Switch for Jarvis model for water uptake reduction
        alphcrit: Optional[float] = _DEPRECATED_ Critical stress index (Jarvis, 1989) for compensation of root water uptake [0.2..1 -, R]
        hlim3h (Optional[float]): Pressure head below which water uptake reduction starts at high Tpot
        hlim3l (Optional[float]): Pressure head below which water uptake reduction starts at low Tpot
        hlim4 (Optional[float]): No water extraction at lower soil water pressure heads
        adcrh (Optional[float]): Level of high atmospheric demand, corresponding to HLIM3H
        adcrl (Optional[float]): Level of low atmospheric demand, corresponding to HLIM3L
        wiltpoint (Optional[float]): Minimum pressure head in leaves
        kstem (Optional[float]): Hydraulic conductance between leaf and root xylem
        rxylem (Optional[float]): Xylem radius
        rootradius (Optional[float]): Root radius
        kroot (Optional[float]): Radial hydraulic conductivity of root tissue
        rootcoefa (Optional[float]): Defines relative distance between roots at which mean soil water content occurs
        swhydrlift (Optional[Literal[0, 1]]): Switch for possibility hydraulic lift in root system
        rooteff (Optional[float]): Root system efficiency factor
        stephr (Optional[float]): Step between values of hroot and hxylem in iteration cycle
        criterhr (Optional[float]): Maximum difference of Hroot between iterations; convergence criterium
        taccur (Optional[float]): Maximum absolute difference between simulated and calculated potential transpiration rate
    """

    swdrought: _Literal[1, 2] | None = None
    swjarvis: _Literal[0, 1, 2, 3, 4] | None = None
    alphcrit: float | None = _Field(default=None, ge=0.2, le=1.0)
    hlim3h: float | None = _Field(default=None, ge=-1.0e4, le=100.0)
    hlim3l: float | None = _Field(default=None, ge=-1.0e4, le=100.0)
    hlim4: float | None = _Field(default=None, ge=-1.6e4, le=100.0)
    adcrh: float | None = _Field(default=None, ge=0.0, le=5.0)
    adcrl: float | None = _Field(default=None, ge=0.0, le=5.0)
    wiltpoint: float | None = _Field(default=None, ge=-1.0e8, le=-1.0e2)
    kstem: float | None = _Field(default=None, ge=1.0e-10, le=10.0)
    rxylem: float | None = _Field(default=None, ge=1.0e-4, le=1.0)
    rootradius: float | None = _Field(default=None, ge=1.0e-4, le=1.0)
    kroot: float | None = _Field(default=None, ge=1.0e-10, le=1.0e10)
    rootcoefa: float | None = _Field(default=None, **_UNITRANGE)
    swhydrlift: _Literal[0, 1] | None = None
    rooteff: float | None = _Field(default=None, **_UNITRANGE)
    stephr: float | None = _Field(default=None, ge=0.0, le=10.0)
    criterhr: float | None = _Field(default=None, ge=0.0, le=10.0)
    taccur: float | None = _Field(default=None, ge=1.0e-5, le=1.0e-2)


class SaltStress(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Salt stress settings for .crp file.

    Attributes:
        swsalinity (Literal[0, 1, 2]): Switch for salt stress

            * 0 - No salt stress
            * 1 - Maas and Hoffman reduction function
            * 2 - Use osmotic head

        saltmax (Optional[float]): Threshold salt concentration in soil water
        saltslope (Optional[float]): Decline of root water uptake above threshold
        salthead (Optional[float]): Conversion factor salt concentration (mg/cm3) into osmotic head (cm)
    """

    swsalinity: _Literal[0, 1, 2] | None = None
    saltmax: float | None = _Field(default=None, ge=0.0, le=100.0)
    saltslope: float | None = _Field(default=None, **_UNITRANGE)
    salthead: float | None = _Field(default=None, ge=0.0, le=1000.0)


class CompensateRWUStress(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Compensate root water uptake stress settings for .crp file.

    Attributes:
        swcompensate (Literal[0, 1, 2]): Switch for compensate root water uptake stress

            * 0 - No compensation
            * 1 - Compensation according to Jarvis (1989)
            * 2 - Compensation according to Walsum (2019)

        swstressor (Optional[Literal[1, 2, 3, 4, 5]]): Switch for stressor

            * 1 - Compensation of all stressors
            * 2 - Compensation of drought stress
            * 3 - Compensation of oxygen stress
            * 4 - Compensation of salinity stress
            * 5 - Compensation of frost stress

        alphacrit (Optional[float]): Critical stress index for compensation of root water uptake
        dcritrtz (Optional[float]): Threshold of rootzone thickness after which compensation occurs
    """

    swcompensate: _Literal[0, 1, 2] | None = None
    swstressor: _Literal[1, 2, 3, 4, 5] | None = None
    alphacrit: float | None = _Field(default=None, ge=0.2, le=1.0)
    dcritrtz: float | None = _Field(default=None, ge=0.02, le=100.0)


class Interception(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Interception settings for .crp file.

    Attributes:
        swinter (Literal[0, 1, 2]): Switch for rainfall interception method

            * 0 - No interception
            * 1 - Agricultural crops (Von Hoyningen-Hune and Braden)
            * 2 - Trees and forests (Gash)

        cofab (Optional[float]): Interception coefficient, corresponding to maximum interception amount
        _table_intertb (Optional[_Table]): _table with the following columns as a function of time T:

            * PFREE - Free throughfall coefficient
            * PSTEM - Stemflow coefficient
            * SCANOPY - Canopy storage coefficient
            * AVPREC = Average rainfall intensity
            * AVEVAP = Average evaporation intensity during rainfall from a wet canopy
    """

    swinter: _Literal[0, 1, 2] | None = None
    cofab: float | None = _Field(default=None, **_UNITRANGE)
    intertb: _Table | None = None


class CO2Correction(
    _PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin, _YAMLUpdateMixin
):
    """CO2 correction settings for WOFOST-type .crp file.

    Attributes:
        swco2 (Literal[0, 1]): Switch for assimilation correction due to CO2 impact

            * 0 - No CO2 assimilation correction
            * 1 - CO2 assimilation correction

        atmofil (Optional[str]): alternative filename for atmosphere.co2
        co2amaxtb (Optional [_Arrays]): Correction of photosynthesis as a function of atmospheric CO2 concentration
        co2efftb (Optional [_Arrays]): orrection of radiation use efficiency as a function of atmospheric CO2 concentration
        co2tratb (Optional [_Arrays]): Correction of transpiration as a function of atmospheric CO2 concentration
    """

    _validation: bool = _PrivateAttr(default=False)
    wofost_variety: _CropVariety | None = _Field(default=None, exclude=True)

    swco2: _Literal[0, 1] | None = None
    atmofil: str | None = None
    co2amaxtb: _Arrays | None = None
    co2efftb: _Arrays | None = None
    co2tratb: _Arrays | None = None


class Preparation(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Preparation, sowing and germination settings for .crp file.

    Attributes:
        swprep (Literal[0, 1]): Switch for preparation.
            * 1 - No preparation.
            * 2 - Preparation before start of crop growth.
        zprep (Optional[float]): Z-level for monitoring work-ability for the crop [-100..0 cm].
        hprep (Optional[float]): Maximum pressure head during preparation [-200..0 cm].
        maxprepdelay (Optional[int]): Maximum delay of preparation from start of growing season [1..366 d].
        swsow (Literal[0, 1]): Switch for sowing
            * 0 - No sowing
            * 1 - Sowing before start of crop growth
        zsow (Optional[float]): Z-level for monitoring work-ability for the crop [-100..0 cm].
        hsow (Optional[float]): Maximum pressure head during sowing [-200..0 cm].
        ztempsow (Optional[float]): Z-level for monitoring temperature for sowing [-100..0 cm].
        tempsow (Optional[float]): Soil temperature needed for sowing [0..30 degrees C].
        maxsowdelay (Optional[int]): Maximum delay of sowing from start of growing season [1..366 d].
        swgerm (Literal[0, 1, 2]): Switch for germination.
            * 0 - No germination.
            * 1 - Germination with temperature sum.
            * 2 - Germination with temperature sum and water potential.
        tsumemeopt (Optional[float]): Temperature sum needed for crop emergence [0..1000 degrees C]
        tbasem (Optional[float]): Minimum temperature, used for germination trajectory [0..40 degrees C].
        teffmx (Optional[float]): Maximum temperature, used for germination trajectory [0..40 degrees C].
        hdrygerm (Optional[float]): Pressure head rootzone for dry germination trajectory [-1000..-0.01 cm].
        hwetgerm (Optional[float]): Pressure head rootzone for wet germination trajectory [-1000..-0.01 cm].
        zgerm (Optional[float]): Z-level for monitoring average pressure head for germination [-100..0 cm].
        agerm (Optional[float]): A-coefficient Eq. 24/25 Feddes & Van Wijk (1988) [0..1000 -].
        swharv (Literal[0, 1]): Switch for harvest.
            * 0 - Timing of harvest depends on end of growing period (CROPEND).
            * 1 - Timing of harvest depends on development stage (DVSEND).
        dvsend (Optional[float]): Development stage at harvest [0..3 -].
    """

    swprep: _Literal[0, 1] | None = _Field(default=None)
    swsow: _Literal[0, 1] | None = None
    swgerm: _Literal[0, 1, 2] | None = None
    swharv: _Literal[0, 1] | None = None
    dvsend: float | None = _Field(default=None, ge=0.0, le=3.0)
    zprep: float | None = _Field(default=None, ge=-100.0, le=0.0)
    hprep: float | None = _Field(default=None, ge=-200.0, le=0.0)
    maxprepdelay: int | None = _Field(default=None, ge=1, le=366)
    zsow: float | None = _Field(default=None, ge=-100.0, le=0.0)
    hsow: float | None = _Field(default=None, ge=-200.0, le=0.0)
    ztempsow: float | None = _Field(default=None, ge=-100.0, le=0.0)
    tempsow: float | None = _Field(default=None, ge=0.0, le=30.0)
    maxsowdelay: int | None = _Field(default=None, ge=1, le=366)
    tsumemeopt: float | None = _Field(default=None, ge=0.0, le=1000.0)
    tbasem: float | None = _Field(default=None, ge=0.0, le=1000.0)
    teffmx: float | None = _Field(default=None, ge=0.0, le=1000.0)
    hdrygerm: float | None = _Field(default=None, ge=-1000.0, le=1000.0)
    hwetgerm: float | None = _Field(default=None, ge=-100.0, le=1000.0)
    zgerm: float | None = _Field(default=None, ge=-100.0, le=1000.0)
    agerm: float | None = _Field(default=None, ge=0.0, le=1000.0)


class GrasslandManagement(
    _PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin, _YAMLUpdateMixin
):
    """Settings specific to the dynamic grass growth module.

    Attributes:
        seqgrazmow (_IntList): sequence of periods with different practices within calender year. Available options:

            * 1 - Grazing
            * 2 - Mowing
            * 3 - Grazing with dewooling

        swharvest (Literal[1, 2]): Switch for timing harvest, either for mowing or grazing

            * 1 - Use dry matter threshold
            * 2 - Use fixed dates

        dateharvest Optional[(_DateList)]: harvest dates (maximum 999)
        swdmgrz Optional[(Literal[1, 2])]: Switch for dry matter threshold to trigger harvest by grazing

            * 1 - Use fixed threshold
            * 2 - Use flexible threshold

        dmgrazing Optional[ _Arrays)]: Minimum dry matter amount for cattle to enter the field [0..1d6 kg DM/ha, R]
        dmgrztb Optional[(int)]: List threshold of above ground dry matter [0..1d6 kg DM/ha, R] to trigger grazing as function of daynumber [1..366 d, R]
        maxdaygrz Optional[(int)]: Maximum growing period after harvest [1..366 -, I]
        swlossgrz Optional[(Literal[0, 1])]: Switch for losses due to insufficient pressure head during grazing

            * 0 - No loss
            * 1 - Losses due to treading

        tagprest Optional[(float)]: Minimum amount of above ground DM after grazing [0..1d6 kg DM/ha, R]
        dewrest Optional[(float)]: Remaining yield above ground after dewooling event [0..1d6 kg DM/ha, R]
        _table_lsda (Optional[_Table]): Actual livestock density of each grazing period
        _table_lsdb (Optional[_Table]): Relation between livestock density, number of grazing days and dry matter uptake
        swdmmow Optional[(int)]: Switch for dry matter threshold to trigger harvest by mowing

            * 1 - Use fixed threshold
            * 2 - Use flexible threshold

        dmharvest Optional[(float)]: Threshold of above ground dry matter to trigger mowing [0..1d6 kg DM/ha, R]
        daylastharvest Optional[(int)]: Last calendar day on which mowing may occur [1..366 -, I]
        dmlastharvest Optional[(float)]: Minimum above ground dry matter for mowing on last date [0..1d6 kg DM/ha, R]
        dmmowtb Optional[(int)]: Dry matter mowing threshold
        maxdaymow Optional[(int)]:Maximum growing period after harvest [1..366 -, I]
        swlossmow Optional[(int)]: Switch for losses due to insufficient pressure head during mowing

            * 0 - No loss
            * 1 - Losses due to treading

        mowrest Optional[(float)]: Remaining yield above ground after mowing event [0..1d6 kg DM/ha, R]
        _table_dmmowdelay Optional[(Optional[_Table])]: Relation between dry matter harvest [0..1d6 kg/ha, R] and days of delay in regrowth [0..366 d, I] after mowing
        swpotrelmf (int): Switch for calculation of potential yield

            * 1 - theoretical potential yield
            * 2 - attainable yield

        relmf (float): Relative management factor to reduce theoretical potential yield to attainable yield [0..1 -, R]
    """

    seqgrazmow: _IntList | None = None
    swharvest: _Literal[1, 2] | None = None
    dateharvest: _Arrays | None = None
    swdmgrz: _Literal[1, 2] | None = None
    dmgrazing: _Decimal2f | None = None
    dmgrztb: _Arrays | None = None
    maxdaygrz: int | None = None
    swlossgrz: _Literal[0, 1] | None = None
    tagprest: _Decimal2f | None = None
    dewrest: _Decimal2f | None = None
    lsda: _Table | None = None
    lsdb: _Table | None = None
    swdmmow: int | None = None
    dmharvest: _Decimal2f | None = None
    daylastharvest: int | None = None
    dmlastharvest: _Decimal2f | None = None
    dmmowtb: _Arrays | None = None
    maxdaymow: int | None = None
    swlossmow: int | None = None
    mowrest: _Decimal2f | None = None
    dmmowdelay: _Table | None = None
    swpotrelmf: int | None = None
    relmf: _Decimal2f | None = None


class CropFile(_PySWAPBaseModel, _FileMixin, _SerializableMixin):
    """Main class for the .crp file.

    This class collects all the settings for the crop file. Currently the types of the
    attributes are set to Any because the validation is not yet implemented.

    Attributes:
        name (str): Name of the crop
        path (Optional[str]): Path to the .crp file
        prep (Optional[Preparation]): Preparation settings
        cropdev_settings (Optional[CropDevelopmentSettings]): Crop development settings
        oxygenstress (Optional[OxygenStress]): Oxygen stress settings
        droughtstress (Optional[DroughtStress]): Drought stress settings
        saltstress (Optional[SaltStress]): Salt stress settings
        compensaterwu (Optional[CompensateRWUStress]): Compensate root water uptake stress settings
        interception (Optional[Interception]): Interception settings
        scheduledirrigation (Optional[ScheduledIrrigation]): Scheduled irrigation settings
        grassland_management (Optional[GrasslandManagement]): Grassland management settings
    """

    _extension: bool = _PrivateAttr(default="crp")

    name: str = _Field(exclude=True)
    path: str | None = None
    prep: _Subsection[Preparation] | None = None
    cropdev_settings: (
        _Subsection[
            CropDevelopmentSettingsFixed
            | CropDevelopmentSettingsWOFOST
            | CropDevelopmentSettingsGrass
        ]
        | None
    ) = None
    oxygenstress: _Subsection[OxygenStress] | None = None
    droughtstress: _Subsection[DroughtStress] | None = None
    saltstress: _Subsection[SaltStress] | None = SaltStress(swsalinity=0)
    compensaterwu: _Subsection[CompensateRWUStress] | None = CompensateRWUStress(
        swcompensate=0
    )
    interception: _Subsection[Interception] | None = None
    scheduledirrigation: _Subsection[_ScheduledIrrigation] | None = (
        _ScheduledIrrigation(schedule=0)
    )
    grasslandmanagement: _Subsection[GrasslandManagement] | None = None
    co2correction: _Subsection[CO2Correction] | None = None

    @property
    def crp(self) -> str:
        """Return the model string of the .crp file.

        The addition validates all the crop files in the dictionary.
        """
        for comp in CropFile.model_fields:
            item = getattr(self, comp)
            if hasattr(item, "validate_with_yaml"):
                item._validation = True
                item.validate_with_yaml()
        return self.model_string()


class Crop(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Crop settings of the simulation.

    Attributes:
        swcrop (int): Switch for crop:

            * 0 - Bare soil.
            * 1 - Simulate crop.

        rds (Optional[float]): Rooting depth of the crop [cm].
        croprotation (Optional[_Table]): _Table with crop rotation data.
        cropfiles (Optional[List[CropFile]]): List of crop files.

    Methods:
        write_crop: Write the crop files.
    """

    swcrop: _Literal[0, 1, None] = None
    rds: float | None = _Field(default=None, ge=1, le=5000)
    croprotation: _Table | None = None
    cropfiles: dict[str, CropFile] = _Field(default_factory=dict, exclude=True)

    def write_crop(self, path: str):
        for name, cropfile in self.cropfiles.items():
            cropfile.save_file(string=cropfile.crp, fname=name, path=path)



# mypy: disable-error-code="attr-defined"
# For some reason, the error was raised on the definition of month labels based
# on pivot table columns (which are dates). Error disabled for now. We will see
# if any bugs come up.

"""
Functions:
    water_content: Plot water content as heatmap.
"""

import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt


def water_content(
    df: pd.DataFrame,
    depth_col: str,
    date_col: str,
    wcontent_col: str,
    title: str = "Water content",
):
    """Plot water content as heatmap.

    For this function to work, the user should either use the `vap` output
    converted to a dataframe, or make sure that in the csv_tz output they
    provide, only the water content data is present.

    Parameters:
        df (pd.DataFrame): DataFrame containing the water content data
        depth_col (str): Column name for depth data
        date_col (str): Column name for date data
        wcontent_col (str): Column name for water content data
        title (str, optional): Title of the plot. Defaults to 'Water content'.
    """

    sns.set_context("poster")

    df_wcont = df[[depth_col, date_col, wcontent_col]]

    df_wcont.loc[:, date_col] = pd.to_datetime(df_wcont[date_col])
    df_wcont.loc[:, depth_col] = df_wcont[depth_col].astype(float)
    df_wcont.loc[:, wcontent_col] = df_wcont[wcontent_col].astype(float)

    pivot_table = df_wcont.pivot(columns=date_col, index=depth_col, values=wcontent_col)
    pivot_table = pivot_table.sort_index(axis=1)

    plt.figure(figsize=(34, 8))
    sns.heatmap(pivot_table, cmap="YlGnBu")
    plt.title(title)
    plt.xlabel("Date")
    plt.ylabel("Depth (cm)")

    plt.gca().invert_yaxis()

    plt.xticks(rotation=45)

    def format_months(x, p):
        return pivot_table.columns[int(x)].strftime("%Y-%m")

    plt.gca().xaxis.set_major_formatter(plt.FuncFormatter(format_months))



# mypy: disable-error-code="call-overload, misc"


from typing import Literal as _Literal

from pydantic import (
    Field as _Field,
    PrivateAttr as _PrivateAttr,
)

from pyswap.components.tables import SOILHYDRFUNC, SOILPROFILE
from pyswap.core.basemodel import PySWAPBaseModel as _PySWAPBaseModel
from pyswap.core.fields import (
    Decimal2f as _Decimal2f,
    Decimal3f as _Decimal3f,
    String as _String,
    Table as _Table,
)
from pyswap.core.valueranges import UNITRANGE as _UNITRANGE
from pyswap.utils.mixins import (
    SerializableMixin as _SerializableMixin,
    YAMLValidatorMixin as _YAMLValidatorMixin,
)

__all__ = [
    "SOILHYDRFUNC",
    "SOILPROFILE",
    "Evaporation",
    "SnowAndFrost",
    "SoilMoisture",
    "SoilProfile",
    "SurfaceFlow",
]


class Evaporation(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Evaporation settings.

    Attributes:
        swcfbs (Literal[0, 1]): Switch for use of soil factor CFBS to calculate Epot from ETref.
        swredu (Literal[0, 1, 2]): Switch for the method for reduction of potential soil evaporation:

            * 0 - reduction to maximum Darcy flux.
            * 1 - reduction to maximum Darcy flux and to maximum Black (1969).
            * 2 - reduction to maximum Darcy flux and to maximum Boesten/Stroosnijder (1986).

        cfevappond (Optional[Decimal2f]): When ETref is used, evaporation coefficient in case of ponding [0..3].
        cfbs (Optional[Decimal2f]): Coefficient for potential soil evaporation [0.5..1.5].
        rsoil (Optional[Decimal2f]): Soil resistance of wet soil [0..1000.0].
        cofredbl (Optional[Decimal2f]): Soil evaporation coefficient of Black [0..1].
        rsigni (Optional[Decimal2f]): Minimum rainfall to reset method of Black [0..100].
        cofredbo (Optional[Decimal2f]): Soil evaporation coefficient of Boesten/Stroosnijder [0..1].
    """

    swcfbs: _Literal[0, 1] | None = None
    swredu: _Literal[0, 1, 2] | None = None
    cfevappond: _Decimal2f | None = _Field(default=None, ge=0, le=3)
    cfbs: _Decimal2f | None = _Field(default=None, ge=0.5, le=1.5)
    rsoil: _Decimal2f | None = _Field(default=None, ge=0, le=1000.0)
    cofredbl: _Decimal2f | None = _Field(default=None, **_UNITRANGE)
    rsigni: _Decimal2f | None = _Field(default=None, ge=0, le=100)
    cofredbo: _Decimal2f | None = _Field(default=None, **_UNITRANGE)


class SnowAndFrost(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Snow and frost settings for the model.

    Attributes:
        swsnow (Literal[0, 1]): Switch for calculation of
            snow accumulation and melt.
        swfrost (Literal[0, 1]): Switch, in case of frost reduce
            soil water flow.
        snowinco (Optional[Decimal2f]): Initial snow water equivalent [0..1000 cm].
        teprrain (Optional[Decimal2f]): Temperature above which all
            precipitation is rain [0..10 oC].
        teprsnow (Optional[Decimal2f]): Temperature below which all
            precipitation is snow [-10..0 oC].
        tfroststa (Optional[Decimal2f]): Soil temperature (oC) where reduction
            of water fluxes starts [-10.0..5.0 oC].
        tfrostend (Optional[Decimal2f]): Soil temperature (oC) where reduction
            of water fluxes ends [-10.0..5.0 oC].
    """

    swsnow: _Literal[0, 1] | None = None
    swfrost: _Literal[0, 1] | None = None
    snowinco: _Decimal2f | None = _Field(default=None, ge=0, le=1000)
    teprrain: _Decimal2f | None = _Field(default=None, ge=0, le=10)
    teprsnow: _Decimal2f | None = _Field(default=None, ge=-10, le=0)
    tfroststa: _Decimal2f | None = _Field(default=None, ge=-10, le=5)
    tfrostend: _Decimal2f | None = _Field(default=None, ge=-10, le=5)


class SoilMoisture(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Soil moisture content and water balance.

    Attributes:
        swinco (Literal[1, 2, 3]): Switch for the type of initial soil moisture condition:

            * 1 - pressure head as function of soil depth.
            * 2 - pressure head of each compartment is in
                hydrostatic equilibrium with initial groundwater level.
            * 3 - read final pressure heads from output file of previous
                Swap simulation.

        head_soildepth (Optional[Table]): Table with head and
            soil depth data.
        gwli (Optional[Decimal2f]): Initial groundwater level [-10000..100 cm].
        inifil (Optional[str]): name of output file *.END which contains
            initial values.
    """

    swinco: _Literal[1, 2, 3] | None = None
    head_soildepth: _Table | None = None
    gwli: _Decimal2f | None = _Field(default=None, ge=-10000, le=100)
    inifil: _String | None = None


class SoilProfile(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Vertical discretization of soil profile, soil hydraulic functions and
        hysteresis of soil water retention.

    Covers parts 4, 5, 6 and 7 of the .swp file.

    Attributes:
        swsophy (Literal[0, 1]): Switch for analytical functions or
            tabular it

            * 0 - Analytical functions with it of Mualem -
                van Genuchten parameters
            * 1 - Soil physical tables

        swhyst (Literal[0, 1, 2]): Hysteresis of soil water retention function

            * 0 - No hysteresis
            * 1 - Hysteresis, initial conditions wetting
            * 2 - Hysteresis, initial conditions drying

        filenamesophy (Optional[str]): Names of it files with
            soil hydraulic tables for each soil layer
        tau (Optional[Decimal2f]): Minimum pressure head difference to change
            wetting-drying [0..1000].
        swmacro (Literal[0, 1]): Switch for preferential flow due to macropores
        soilprofile (Table): Table with soil profile data
        soilhydrfunc (Optional[Table]): Table with
            soil hydraulic functions
    """

    _validation: bool = _PrivateAttr(default=False)

    swsophy: _Literal[0, 1] | None = None
    swhyst: _Literal[0, 1, 2] | None = None
    swmacro: _Literal[0, 1] | None = None
    filenamesophy: _String | None = None
    tau: _Decimal2f | None = _Field(default=None, ge=0, le=1000)
    soilprofile: _Table | None = None
    soilhydrfunc: _Table | None = None


class SurfaceFlow(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Surface flow settings (ponding, runoff and runon).

    Attributes:
        swpondmx (Literal[0, 1]): Switch for variation ponding
            threshold for runoff

            * 0 - Ponding threshold for runoff is constant
            * 1 - Ponding threshold for runoff varies in time

        swrunon (Literal[0, 1]): Switch for runon

            * 0 - No runon
            * 1 - Use runon data

        rsro (Optional[Decimal2f]): Drainage resistance for surface runoff [0.001..1.0].
        rsroexp (Optional[Decimal2f]): Exponent for drainage equation of surface runoff [0.01..10.0].
        pondmx (Optional[Decimal2f]): In case of ponding, minimum thickness for runoff [0..1000].
        rufil (Optional[str]): Name of the runon file.
        pondmxtb (Optional[Table]): Minimum thickness for runoff as a function of time.
    """

    swpondmx: _Literal[0, 1] | None = None
    swrunon: _Literal[0, 1] | None = None
    rsro: _Decimal3f | None = _Field(default=None, ge=0.001, le=1.0)
    rsroexp: _Decimal2f | None = _Field(default=None, ge=0.01, le=10.0)
    pondmx: _Decimal2f | None = _Field(default=None, ge=0, le=1000)
    rufil: _String | None = None
    pondmxtb: _Table | None = None



"""Plot evapotranspiration (potential vs actual) and compute the RMSE."""

import matplotlib.pyplot as plt
import seaborn as sns
from pandas import DataFrame


def evapotranspiration(
    potential: DataFrame, actual: DataFrame, title: str = "Evapotranspiration"
):
    """Plot evapotranspiration (potential vs actual) and compute the RMSE.

    Paremeters:
        potential (DataFrame): DataFrame containing dates and values for potential evapotranspiration.
        actual (DataFrame): DataFrame containing dates and values for actual evapotranspiration.
        title (str, optional): Title of the plot. Defaults to 'Evapotranspiration'.
    """

    sns.set_context("poster")

    _fig, ax = plt.subplots(figsize=(34, 8))
    sns.lineplot(data=potential, ax=ax, label="Potential", color="black", linewidth=1)
    sns.lineplot(
        data=actual, ax=ax, label="Actual", color="orange", linewidth=1, linestyle="--"
    )

    ax.set_title(title, pad=20)
    ax.set_xlabel("Date")
    ax.set_ylabel("Evapotranspiration")

    ax.tick_params(axis="x", rotation=45)
    ax.legend()
    plt.tight_layout()
    plt.show()



from pandas import DataFrame, to_datetime
from pydantic import BaseModel as _BaseModel

from pyswap.core.io.io_csv import load_csv
from pyswap.libs import co2concentration
from pyswap.utils.mixins import FileMixin as _FileMixin


class CO2concentration(_BaseModel, _FileMixin):
    path: str = co2concentration
    _extension: str = "co2"

    def get_filtered_co2file(self, period) -> DataFrame:
        """Get CO2 concentration data for a given period.

        Parameters
        ----------
        period : list
            List of two strings representing the start and end date (e.g. ["2000-01-01", "2020-12-31"]).
        """
        data = load_csv(
            self.path,
            delimiter=" ",
            skiprows=12,
            index_col=0,
        )

        # Convert period to datetime
        period = to_datetime(period)

        # Filter data
        data = data[(data.index >= period[0].year) & (data.index <= period[1].year)]

        return data

    def write_co2(self, directory, period) -> None:
        """Write CO2 concentration data to a file."""
        # Get data for the given period
        data = self.get_filtered_co2file(period)

        # Create datastring
        dstring = data.to_csv(
            index=True,
            float_format="%.2f",
            sep=" ",
            lineterminator="\n",
        )

        # Write to file
        self.save_file(string=dstring, fname="Atmospheric", path=directory)



"""Custom field types for pySWAP serialization and deserialization.

The use of custom Annotated fields in combination with Pydantic allows for the
serialization and deserialization of objects to the appropriate format used in
the SWAP. This is particularily useful when Should a new type of field be
implemented, the design should follow the pattern below:

```Python
CustonType = Annotated[
    <native python type>,
    pydantic.AfterValidator(custom_parsing_function),
    pydantic.PlainSerializer(custom_serializer, return_type=<native python type>, when_used="json"),
    pydantic.Field(<custom field parameters>)
    ]
```

Where:

- <native python type>: A native python type of the field that he user has to
    provide (e.g., int).
- pydantic.AfterValidator (optional):
    Pydantic validator runs after the field value passed by the user is
    initially parsed and validated (e.g., for required fields). It then takes a
    custom parsing function as an argument and returns the final value
    compatible with pySWAP serialization. For custom validation functions refer
    to `pyswap.core.parsers` module.
- pydantic.PlainSerializer:
    A serializer in called when Model.model_dump(mode="json") is called.
    Internally in pySWAP this is done in PySWAPBaseModel.model_string() method.
    It takes a custom serialization function as an argument and returns the
    parameter in a string format compatible with SWAP. For custom serializer
    functions, refer to `pyswap.core.serializers` module.
- pydantic.Field:
    In Pydantic Field object is used to define metadata for a field. In pySWAP,
    the context attribute is used to pass additional information to the
    serializer functions.

Then use the custom type in the models:

```Python
class SWAPSection(PySWAPBaseModel):
    field: CustomType
```

Fields in this module:

    Table (DataFrame): A DataFrame object serialized as a string with just
        the headers and the data.
    Arrays (DataFrame): A DataFrame object serialized as a string with just
        the columns of data (no headers),
        but with the variable name in front (e.g., FLUXTB = 0.0 0.0/n 1.0 1.0 )
    CSVTable (DataFrame): A DataFrame object serialized as a string with
        the headers and data in CSV format,
        specifically tailored for the .met file format.
    DayMonth (d): A date object serialized as a string with just the day and
        month (e.g., '01 01').
    StringList (List[str]): A list of strings serialized as a string with
        the elements separated by commas, enclosed
        in quotation marks (e.g., 'string1, string2, string3').
    FloatList (List[float]): A list of floats serialized as a string with
        the elements separated by spaces.
    DateList (List[d]): A list of date objects serialized as a string with
        the elements separated by newlines.
    Switch (bool | int): A boolean or integer serialized as
        an integer (0 or 1).
    ObjectList (list): A list of objects serialized as a string with
        the elements separated by newlines.
"""

from datetime import date
from decimal import Decimal
from typing import Annotated, TypeVar

from pandas import DataFrame
from pydantic import AfterValidator, BeforeValidator, Field
from pydantic.functional_serializers import PlainSerializer

from pyswap.core.basemodel import PySWAPBaseModel
from pyswap.core.parsers import (
    parse_day_month,
    parse_decimal,
    parse_float_list,
    parse_int_list,
    parse_quoted_string,
    parse_string_list,
)
from pyswap.core.serializers import (
    serialize_arrays,
    serialize_day_month,
    serialize_decimal,
    serialize_table,
)

__all__ = [
    "Arrays",
    "DateList",
    "DayMonth",
    "Decimal2f",
    "Decimal3f",
    "Decimal4f",
    "File",
    "FloatList",
    "IntList",
    "String",
    "StringList",
    "Subsection",
    "Table",
]

Table = Annotated[
    DataFrame,
    PlainSerializer(serialize_table, return_type=str, when_used="json"),
    Field(json_schema_extra={"is_annotated_exception_type": True}),
]
"""Serialize pd.DataFrame with headers to a string without leading variable name."""

Arrays = Annotated[
    DataFrame, PlainSerializer(serialize_arrays, return_type=str, when_used="json")
]
"""Serialize pd.DataFrame without headers to a string with leading variable name."""

DayMonth = Annotated[
    date | str,
    AfterValidator(parse_day_month),
    PlainSerializer(serialize_day_month, return_type=str, when_used="json"),
]
"""Serialize date object to a string with just the day and month."""

StringList = Annotated[
    list[str] | str,
    AfterValidator(parse_string_list),
    PlainSerializer(lambda x: f"'{','.join(x)}'", return_type=str, when_used="json"),
]
"""Serialize list of strings to a string with elements separated by commas."""

FloatList = Annotated[
    list[float] | str,
    AfterValidator(parse_float_list),
    PlainSerializer(
        lambda x: " ".join([f"{Decimal(f):.2f}" for f in x]),
        return_type=str,
        when_used="json",
    ),
]
"""Serialize list of floats to a string with elements separated by spaces."""

IntList = Annotated[
    list[int] | str,
    AfterValidator(parse_int_list),
    PlainSerializer(
        lambda x: " ".join([str(f) for f in x]), return_type=str, when_used="json"
    ),
]
"""Serialize list of integers to a string with elements separated by spaces."""

DateList = Annotated[
    list[date],
    PlainSerializer(
        lambda x: "\n" + "\n".join([d.strftime("%Y-%m-%d") for d in x]),
        return_type=str,
        when_used="json",
    ),
]
"""Serialize list of date objects to a string with elements separated by newlines."""

String = Annotated[
    str,
    PlainSerializer(lambda x: f"'{x}'", return_type=str),
    AfterValidator(parse_quoted_string),
]
"""Serialize string to a string with quotation marks."""

File = Annotated[
    PySWAPBaseModel,
    PlainSerializer(lambda x: x.model_string(), return_type=str, when_used="json"),
    Field(json_schema_extra={"is_annotated_exception_type": True}),
]
"""Serialize PySWAPBaseModel to a string."""

SubsectionTypeVar = TypeVar("SubsectionTypeVar", bound=PySWAPBaseModel)

Subsection = Annotated[
    SubsectionTypeVar,
    PlainSerializer(lambda x: x.model_string(), return_type=str),
    Field(json_schema_extra={"is_annotated_exception_type": True}),
]
"""Serialize a nested PySWAPBaseModel to a string."""

Decimal2f = Annotated[
    float | str,
    BeforeValidator(parse_decimal),
    PlainSerializer(serialize_decimal(precision=2), return_type=str, when_used="json"),
]
"""Serialize float to a string with 2 decimal places."""

Decimal3f = Annotated[
    float,
    PlainSerializer(serialize_decimal(precision=3), return_type=str, when_used="json"),
]
"""Serialize float to a string with 3 decimal places."""

Decimal4f = Annotated[
    float,
    PlainSerializer(serialize_decimal(precision=4), return_type=str, when_used="json"),
]
"""Serialize float to a string with 4 decimal places."""



# mypy: disable-error-code="call-overload, misc"

"""Lateral drainage settings

Settings for the lateral drainage of the .swp file, including the .dra file settings.

Classes:
    Flux: Fluxes between drainage levels in .dra file.
    DraFile: Drainage file (.dra) settings.
    Drainage: The lateral drainage settings of .swp file.
"""

from typing import Literal as _Literal

from pydantic import (
    Field as _Field,
    PrivateAttr as _PrivateAttr,
)

from pyswap.components.tables import (
    DATOWLTB1,
    DATOWLTB2,
    DATOWLTB3,
    DATOWLTB4,
    DATOWLTB5,
    DRAINAGELEVELTOPPARAMS,
    DRNTB,
    MANSECWATLVL,
    PRIWATLVL,
    QDRNTB,
    QWEIR,
    QWEIRTB,
    SECWATLVL,
)
from pyswap.core.basemodel import PySWAPBaseModel as _PySWAPBaseModel
from pyswap.core.defaults import FNAME_IN as _FNAME_IN
from pyswap.core.fields import (
    File as _File,
    FloatList as _FloatList,
    String as _String,
    Subsection as _Subsection,
    Table as _Table,
)
from pyswap.core.valueranges import UNITRANGE as _UNITRANGE
from pyswap.utils.mixins import (
    FileMixin as _FileMixin,
    SerializableMixin as _SerializableMixin,
    YAMLValidatorMixin as _YAMLValidatorMixin,
)

__all__ = [
    "DATOWLTB1",
    "DATOWLTB2",
    "DATOWLTB3",
    "DATOWLTB4",
    "DATOWLTB5",
    "DRAINAGELEVELTOPPARAMS",
    "DRNTB",
    "MANSECWATLVL",
    "PRIWATLVL",
    "QDRNTB",
    "QWEIR",
    "QWEIRTB",
    "SECWATLVL",
    "DraFile",
    "Drainage",
    "Flux",
]


class Flux(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """Fluxes between drainage levels in .dra file.

    !!! note

        This was rewritten to be a single class instead of a list of classes.
        Simplicity over DRY. Anyway, the prefered way to set this up would be
        through the table from the extended section I guess.

    Attributes:
        drares (float): Drainage resistance [10 .. 1e5 d].
        infres (float): Infiltration resistance [10 .. 1e5 d].
        swallo (Literal[1, 2]): Switch to allow drainage from this level.

            * 1 - Drainage and infiltration are both allowed.
            * 2 - Only infiltration is allowed.
            * 3 - Only drainage is allowed.

        l (Optional[float]): Drain spacing [1 .. 1e5 m].
        zbotdr (float): Level of the bottom of the drain [-1e4 .. 0 cm].
        swdtyp (Literal[1, 2]): Drainage type.

            * 1 - drain tube.
            * 2 - open channel.

        datowltb (Table): date DATOWL [date] and channel water
            level LEVEL. Add suffix to the dataframe headers
            according to the level number.
    """

    drares1: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    infres1: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    swallo1: _Literal[1, 2, 3] | None = None
    l1: float | None = _Field(default=None, ge=1.0, le=1.0e5)
    zbotdr1: float = _Field(default=None, ge=-1000.0, le=0.0)
    swdtyp1: _Literal[1, 2] | None = None
    datowltb1: _Table | None = None
    # level 2
    drares2: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    infres2: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    swallo2: _Literal[1, 2, 3] | None = None
    l2: float | None = _Field(default=None, ge=1.0, le=1.0e5)
    zbotdr2: float | None = _Field(default=None, ge=-1000.0, le=0.0)
    swdtyp2: _Literal[1, 2] | None = None
    datowltb2: _Table | None = None
    # level 3
    drares3: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    infres3: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    swallo3: _Literal[1, 2, 3] | None = None
    l3: float | None = _Field(default=None, ge=1.0, le=1.0e5)
    zbotdr3: float | None = _Field(default=None, ge=-1000.0, le=0.0)
    swdtyp3: _Literal[1, 2] | None = None
    datowltb3: _Table | None = None
    # level 4
    drares4: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    infres4: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    swallo4: _Literal[1, 2, 3] | None = None
    l4: float | None = _Field(default=None, ge=1.0, le=1.0e5)
    zbotdr4: float | None = _Field(default=None, ge=-1000.0, le=0.0)
    swdtyp4: _Literal[1, 2] | None = None
    datowltb4: _Table | None = None
    # level 5
    drares5: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    infres5: float | None = _Field(default=None, ge=10.0, le=1.0e5)
    swallo5: _Literal[1, 2, 3] | None = None
    l5: float | None = _Field(default=None, ge=1.0, le=1.0e5)
    zbotdr5: float | None = _Field(default=None, ge=-1000.0, le=0.0)
    swdtyp5: _Literal[1, 2] | None = None
    datowltb5: _Table | None = None


class DraFile(_PySWAPBaseModel, _FileMixin, _SerializableMixin):
    """Content of the drainage file (.dra).

    Attributes general:
        dramet (Literal[1, 2, 3]): Method of lateral drainage calculation

            * 1 - Use table of drainage flux - groundwater level relation.
            * 2 - Use drainage formula of Hooghoudt or Ernst.
            * 3 - Use drainage/infiltration resistance, multi-level if needed.

        swdivd (Literal[1, 2]): Calculate vertical distribution of
            drainage flux in groundwater.
        cofani (Optional[FloatList]): specify anisotropy factor COFANI
            (horizontal/vertical saturated hydraulic conductivity) for
            each soil layer (maximum MAHO)
        swdislay (Literal[0, 1, 2, 3, '-']): Switch to adjust
            upper boundary of model discharge layer.

            * 0 - No adjustment
            * 1 - Adjustment based on depth of top of model discharge
            * 2 - Adjustment based on factor of top of model discharge

    Attributes drainage flux table (option 1):
        lm1 (float): Drain spacing
        table_qdrntb (Table): Table of drainage flux - groundwater level.

    Attributes drainage formula (option 2):
        lm2 (float): Drain spacing.
        shape (float): Shape factor to account for actual location between
            drain and water divide.
        wetper (float): Wet perimeter of the drain.
        zbotdr (float): Level of drain bottom.
        entres (float): Drain entry resistance.
        ipos (Literal[1, 2, 3, 4, 5]): Position of drain

            * 1 - On top of an impervious layer in a homogeneous profile
            * 2 - Above an impervious layer in a homogeneous profile
            * 3 - At the interface of a fine upper and a coarse lower
                soil layer
            * 4 - In the lower, more coarse soil layer
            * 5 - In the upper, more fine soil layer

        basegw (float): Level of impervious layer.
        khtop (float): Horizontal hydraulic conductivity of the top layer.
        khbot (Optional[float]): Horizontal hydraulic conductivity of
            the bottom layer.
        zintf (Optional[float]): Interface level of the coarse and
            fine soil layer.
        kvtop (Optional[float]): Vertical hydraulic conductivity of
            the top layer.
        kvbot (Optional[float]): Vertical hydraulic conductivity of
            the bottom layer.
        geofac (Optional[float]): Geometric factor of Ernst.

    Attributes drainage infiltration resistance (option 3):
        nrlevs (int): Number of drainage levels.
        swintfl (Literal[0, 1]): Option for interflow in highest
            drainage level (shallow system with short residence time).
        cofintflb (float): Coefficient for interflow relation.
        expintflb (float): Exponent for interflow relation.
        swtopnrsrf (Literal[0, 1]): Switch to enable adjustment of
            model discharge layer.
        fluxes (Flux): Flux object containing parameters for each drainage level.

    Attributes extended section (surface water management):
        altcu (float): Altitude of the control unit relative to reference level.
        nrsrf (int): Number of subsurface drainage levels.
        swnrsrf (Literal[0, 1, 2]): Switch to introduce rapid subsurface drainage.
        rsurfdeep (Optional[float]): Maximum resistance of rapid subsurface drainage.
        rsurfshallow (Optional[float]): Minimum resistance of rapid subsurface drainage.
        swsrf (Literal[1, 2, 3]): Switch for interaction with surface water system.
        swsec (Optional[Literal[1, 2]]): Option for surface water level of secondary system.
        wlact (Optional[float]): Initial surface water level.
        osswlm (Optional[float]): Criterium for warning about oscillation.
        nmper (Optional[int]): Number of management periods.
        swqhr (Optional[Literal[1, 2]]): Switch for type of discharge relationship.
        sofcu (Optional[float]): Size of the control unit.
    """

    _extension = _PrivateAttr("dra")
    # General
    dramet: _Literal[1, 2, 3] | None = None
    swdivd: _Literal[1, 2] | None = None
    cofani: _FloatList | None = None
    swdislay: _Literal[0, 1, 2, 3, "-"] | None = None
    # Drainage flux table
    lm1: float | None = _Field(default=None, ge=1.0, le=1000.0)
    qdrntb: _Table | None = None
    # Drainage formula
    lm2: float | None = _Field(default=None, ge=1.0, le=1000.0)
    shape: float | None = _Field(default=None, **_UNITRANGE)
    wetper: float | None = _Field(default=None, ge=0.0, le=1000.0)
    zbotdr: float | None = _Field(default=None, ge=-1000.0, le=0.0)
    entres: float | None = _Field(default=None, ge=0.0, le=1000.0)
    ipos: _Literal[1, 2, 3, 4, 5] | None = None
    basegw: float | None = _Field(default=None, ge=-1.0e4, le=0.0)
    khtop: float | None = _Field(default=None, ge=0.0, le=1000.0)
    khbot: float | None = _Field(default=None, ge=0.0, le=1000.0)
    zintf: float | None = _Field(default=None, ge=-1.0e4, le=0.0)
    kvtop: float | None = _Field(default=None, ge=0.0, le=1000.0)
    kvbot: float | None = _Field(default=None, ge=0.0, le=1000.0)
    geofac: float | None = _Field(default=None, ge=0.0, le=100.0)
    # Drainage infiltration resistance
    nrlevs: int | None = _Field(default=None, ge=1, le=5)
    swintfl: _Literal[0, 1] | None = None
    cofintflb: float | None = _Field(default=None, ge=0.01, le=10.0)
    expintflb: float | None = _Field(default=None, ge=0.1, le=1.0)
    swtopnrsrf: _Literal[0, 1] | None = None
    fluxes: _Subsection | None = None
    # Extended section
    altcu: float | None = _Field(default=None, ge=-300000.0, le=300000.0)
    drntb: _Table | None = None
    nrsrf: int | None = _Field(default=None, ge=1, le=5)
    swnrsrf: _Literal[0, 1, 2] | None = None
    rsurfdeep: float | None = _Field(default=None, ge=0.001, le=1000.0)
    rsurfshallow: float | None = _Field(default=None, ge=0.001, le=1000.0)
    cofintfl: float | None = _Field(default=None, ge=0.01, le=10.0)
    expintfl: float | None = _Field(default=None, ge=0.01, le=10.0)
    swsrf: _Literal[1, 2, 3] | None = None
    swsec: _Literal[1, 2] | None = None
    secwatlvl: _Table | None = None
    wlact: float | None = _Field(default=None, ge=-300000.0, le=300000.0)
    osswlm: float | None = _Field(default=None, ge=0.0, le=10.0)
    nmper: int | None = _Field(default=None, ge=1, le=3660)
    swqhr: _Literal[1, 2] | None = None
    sofcu: float | None = _Field(default=None, ge=0.1, le=100000.0)
    mansecwatlvl: _Table | None = None
    drainageleveltopparams: _Table | None = None
    qweir: _Table | None = None
    qweirtb: _Table | None = None
    priwatlvl: _Table | None = None

    @property
    def dra(self):
        return self.model_string()


class Drainage(_PySWAPBaseModel, _SerializableMixin, _YAMLValidatorMixin):
    """The lateral drainage settings inside .swp file.

    Attributes:
        swdra (Literal[0, 1, 2]): Switch for lateral drainage.

            * 0 - No drainage.
            * 1 - Simulate with a basic drainage routine.
            * 2 - Simulate with surface water management.

        drfil (str): Name of the file. This attribute is frozen, there is no
            need to change it.
        drafile (Optional[Any]): Content of the drainage file.
    """

    swdra: _Literal[0, 1, 2] | None = None
    drfil: _String | None = _Field(default=_FNAME_IN, frozen=True)
    drafile: _File | None = _Field(default=None, exclude=True)

    def write_dra(self, path: str) -> None:
        self.drafile.save_file(string=self.drafile.dra, fname=self.drfil, path=path)
        return None



"""
SWAP executable management module.

This module provides functionality to download and manage SWAP executables
similar to how flopy handles MODFLOW executables.
"""

import os
import shutil
import subprocess
import urllib.request
from pathlib import Path

import typer
import yaml

import pyswap
from pyswap.core.defaults import IS_WINDOWS


def get_swap_executable_path() -> Path:
    """Get the path where SWAP executable should be installed."""
    # Use the package installation directory
    package_dir = Path(pyswap.__file__).parent
    exe_dir = package_dir / "libs" / "swap"

    # make a directory if it doesn't exist
    exe_dir.mkdir(parents=True, exist_ok=True)

    exe_path = exe_dir / ("swap.exe" if IS_WINDOWS else "swap420")

    return exe_path


def write_version_info(version: str):
    version_file = get_swap_executable_path().parent / "version.yaml"
    version_data = {"version_swap": version}
    with version_file.open("w") as f:
        yaml.dump(version_data, f)


def read_swap_version() -> str | None:
    """Read the SWAP version from the version file."""
    version_file = get_swap_executable_path().parent / "version.yaml"
    if version_file.exists():
        try:
            with version_file.open("r") as f:
                data = yaml.safe_load(f)
                return data.get("version_swap")  # type: ignore[no-any-return]
        except Exception:
            return None
    return None


def show_info() -> dict:
    """Display information about pySWAP and SWAP versions and installation.

    Returns:
        Information dictionary with version and status details.

    Examples:
        >>> import pyswap as psp
        >>> info = psp.show_info()
        pySWAP version: 0.2.8
        SWAP version: 4.2.0
        SWAP executable: Available at /path/to/swap420

        >>> # Get info without printing
        >>> info = psp.show_info(verbose=False)
        >>> print(f"Running pySWAP {info['pyswap_version']}")
    """
    import pyswap

    # Get pySWAP version
    pyswap_version = pyswap.__version__

    # Get SWAP version and executable info
    swap_version = read_swap_version()
    swap_exe_path = get_swap_executable_path()
    swap_available = swap_exe_path.exists() and os.access(swap_exe_path, os.X_OK)

    # Get platform info
    system = "Windows" if IS_WINDOWS else "Linux"

    info = {
        "pyswap_version": pyswap_version,
        "swap_version": swap_version,
        "swap_executable_path": str(swap_exe_path),
        "swap_available": swap_available,
        "platform": system,
        "package_location": str(Path(pyswap.__file__).parent),
    }

    typer.echo("=" * 40)
    typer.echo("pySWAP Setup Information:")
    typer.echo(f"  pySWAP version: {pyswap_version}")
    if swap_version:
        version_label = (
            "custom/uploaded" if swap_version == "custom" else swap_version
        )
        typer.echo(f"  SWAP version: {version_label}")
    else:
        typer.echo("  SWAP version: Not installed")

    if swap_available:
        typer.echo(f"  SWAP executable: ✓ Available at {swap_exe_path}")
    else:
        typer.echo(f"  SWAP executable: ✗ Not found at {swap_exe_path}")

    typer.echo(f"  Platform: {system}")
    typer.echo(f"  Package location: {info['package_location']}")

    if not swap_available:
        typer.echo("")
        typer.echo("To install SWAP:")
        typer.echo("  - Download: pyswap get-swap")
        typer.echo("  - Upload local file: pyswap upload-swap <path-to-executable>")

    typer.echo("=" * 40)

    return info


def get_swap(
    version: str = "4.2.0",
    force: bool = False,
    verbose: bool = True,
    auto_install: bool = False,
) -> str:
    """Download and setup SWAP executable for the current platform.

    Args:
        version: SWAP version to download. Defaults to "4.2.0".
        force: Force re-download even if executable already exists. Defaults to False.
        verbose: Print download progress and information. Defaults to True.
        auto_install: Automatically install SWAP if not found during the model run. Defaults to False.
    Returns:
        Path to the SWAP executable.

    Examples:
        >>> import pyswap as psp
        >>> swap_exe = psp.get_swap()
        >>> print(f"SWAP executable available at: {swap_exe}")

        >>> # Force re-download latest version
        >>> swap_exe = psp.get_swap(force=True, verbose=True)
    """

    exe_path = get_swap_executable_path()

    # Check if executable already exists
    if os.path.exists(exe_path) and not force:
        if verbose:
            typer.echo(f"SWAP executable already exists at: {exe_path}")
            typer.echo("Use force=True to re-download")
        return exe_path

    # Download and install
    if verbose or auto_install:
        platform_name = "windows" if IS_WINDOWS else "linux"
        typer.echo(f"Downloading SWAP {version} for {platform_name}...")
        typer.echo(f"Installing to: {exe_path.parent}")

    try:
        _download_swap_executable(
            version=version, target_path=exe_path, verbose=verbose or auto_install
        )

        write_version_info(version)
        if verbose or auto_install:
            typer.echo(f"✓ SWAP {version} successfully installed!")
            typer.echo(f"Executable path: {exe_path}")
            typer.echo("Updated version info file.")
    except Exception as e:
        msg = f"Failed to download SWAP executable: {e}"
        raise typer.Exit(msg) from e
    else:
        return exe_path


def _download_swap_executable(
    version: str, target_path: str, verbose: bool = True
) -> None:
    """Download and extract SWAP executable from GitHub releases."""

    # Construct download URL
    base_url = "https://github.com/SWAP-model/SWAP/releases/download"
    platform_name = "windows" if IS_WINDOWS else "linux"
    filename = f"swap{version}-{platform_name}"
    if IS_WINDOWS:
        filename += ".exe"

    url = f"{base_url}/v{version}/{filename}"

    if verbose:
        typer.echo(f"Downloading from: {url}")

    try:
        # Download with progress
        if verbose:
            with typer.progressbar(length=100, label="Downloading") as progress:  # type: ignore[var-annotated]

                def progress_hook(
                    block_num: int, block_size: int, total_size: int
                ) -> None:
                    downloaded = block_num * block_size
                    if total_size > 0:
                        percent = min(100, (downloaded * 100) // total_size)
                        progress.update(1 if percent > progress.pos else 0)

                urllib.request.urlretrieve(url, target_path, progress_hook)
        else:
            urllib.request.urlretrieve(url, target_path)

    except urllib.error.HTTPError as e:
        if e.code == 404:
            platform_name = "windows" if IS_WINDOWS else "linux"
            msg = (
                f"SWAP version {version} not found for platform {platform_name}. "
                f"Check available releases at: https://github.com/SWAP-model/SWAP/releases"
            )
            raise RuntimeError(msg) from e
        else:
            msg = f"Download failed: HTTP {e.code}"
            raise RuntimeError(msg) from e
    except Exception as e:
        msg = f"Download failed: {e}"
        raise RuntimeError(msg) from e

    # Make executable (Unix systems)
    if not IS_WINDOWS:
        os.chmod(target_path, 0o755)


def check_swap(exe_path: str | None = None, verbose: bool = True) -> bool:
    """Check if SWAP executable is available and working by running a testcase.

    Args:
        exe_path: Path to SWAP executable. If None, uses default package location.
        verbose: Print information about the check. Defaults to True.

    Returns:
        True if SWAP is available and working.
    """
    if exe_path is None:
        exe_path = str(get_swap_executable_path())

    if not os.path.exists(exe_path):
        if verbose:
            typer.echo(f"SWAP executable not found at: {exe_path}")
            typer.echo("Run get_swap() to download and install SWAP")
        return False

    # Try to run SWAP by executing a simple testcase
    try:
        from pyswap import testcase

        if verbose:
            show_info()
            typer.echo(f"Running testcase to verify SWAP executable: {exe_path}")
            typer.echo("Loading hupselbrook testcase...")

        # Get the testcase and run it
        hupselbrook = testcase.get("hupselbrook")
        result = hupselbrook.run()
        print(result)

        if verbose:
            typer.echo("✓ SWAP testcase completed successfully!")
            typer.echo(f"✓ SWAP executable working: {exe_path}")
            # Show a brief summary
            yearly_data = result.yearly_summary
            if not yearly_data.empty:
                total_years = len(yearly_data)
                avg_rain = yearly_data['RAIN'].mean()
                avg_drainage = yearly_data['DRAINAGE'].mean()
                typer.echo(f"Testcase summary: {total_years} years, avg rainfall: {avg_rain:.1f}mm, avg drainage: {avg_drainage:.1f}mm")

    except Exception as e:
        if verbose:
            typer.echo(f"✗ Error running SWAP testcase: {e}")
            typer.echo("This could indicate a problem with the SWAP executable or installation")
        return False
    else:
        return True
def upload_swap(
    file_path: str,
    version: str,
    force: bool = False,
    verbose: bool = True,
) -> str:
    """Install SWAP executable from a local file.

    Args:
        file_path: Path to the local SWAP executable file to install.
        version: Version identifier for the uploaded executable. Defaults to "custom".
        force: Force replace existing executable. Defaults to False.
        verbose: Print installation progress and information. Defaults to True.

    Returns:
        Path to the installed SWAP executable.

    Examples:
        >>> import pyswap as psp
        >>> # Install from a local executable
        >>> swap_exe = psp.upload_swap("/path/to/my/swap", version="4.2.1-custom")
        >>> print(f"SWAP executable installed at: {swap_exe}")

        >>> # Force replace existing installation
        >>> swap_exe = psp.upload_swap("./swap420", force=True, verbose=True)
    """
    from pathlib import Path

    source_path = Path(file_path)
    target_path = get_swap_executable_path()

    # Validate source file
    if not source_path.exists():
        msg = f"Source file not found: {source_path}"
        raise FileNotFoundError(msg)

    if not source_path.is_file():
        msg = f"Source path is not a file: {source_path}"
        raise ValueError(msg)

    # Check if target already exists
    if target_path.exists() and not force:
        current_version = read_swap_version()
        msg = (
            f"SWAP executable already exists at: {target_path}\n"
            f"Current version: {current_version or 'unknown'}\n"
            f"Use force=True to replace it."
        )
        raise FileExistsError(msg)

    if verbose:
        typer.echo(f"Installing SWAP executable from: {source_path}")
        typer.echo(f"Target location: {target_path}")
        if target_path.exists():
            typer.echo("Replacing existing executable...")

    try:
        # Create target directory if it doesn't exist
        target_path.parent.mkdir(parents=True, exist_ok=True)

        # Copy the file
        shutil.copy2(source_path, target_path)

        # Make executable (Unix systems)
        if not IS_WINDOWS:
            os.chmod(target_path, 0o755)

        # Verify the copied file
        if not _verify_executable(target_path, verbose=verbose):
            # Clean up on failure
            target_path.unlink(missing_ok=True)
            msg = "Uploaded file does not appear to be a valid executable"
            raise RuntimeError(msg)

        # Write version info
        write_version_info(version)

        if verbose:
            typer.echo("✓ Your SWAP executable successfully installed!")
            typer.echo(f"Version: {version}")
            typer.echo(f"Executable path: {target_path}")

        return str(target_path)

    except Exception as e:
        if verbose:
            typer.echo(f"✗ Failed to install SWAP executable: {e}")
        raise


def _verify_executable(exe_path: Path, verbose: bool = False) -> bool:
    """Verify that the uploaded file is a valid executable.

    This could be expanded if we could make sure there --version always works on swap.
    For now it seems it spits out the version output to a file instead of just stdout.

    Args:
        exe_path: Path to the executable to verify.
        verbose: Print verification details.

    Returns:
        True if the file appears to be a valid executable.
    """
    try:
        # Check if file is executable
        if not os.access(exe_path, os.X_OK):
            if verbose:
                typer.echo("File is not executable")
            return False

        typer.echo("Warning: Uploading unverified executable.")

    except subprocess.TimeoutExpired:
        if verbose:
            typer.echo("Warning: Executable test timed out, but installing anyway")
        return True
    except Exception as e:
        if verbose:
            typer.echo(
                f"Warning: Could not verify executable ({e}), but installing anyway"
            )
        return True
    else:
        return True


def remove_swap(verbose: bool = True) -> bool:
    """Remove the SWAP executable from the package directory.

    Args:
        verbose: Print information about the removal. Defaults to True.

    Returns:
        True if removal was successful.
    """
    exe_path = get_swap_executable_path()

    if not os.path.exists(exe_path):
        if verbose:
            typer.echo("SWAP executable not found, nothing to remove")
        return True

    try:
        os.remove(exe_path)
        os.remove(exe_path.parent / "version.yaml")
        if verbose:
            typer.echo(f"✓ SWAP executable removed from: {exe_path}")
            typer.echo(
                f"✓ version.yaml removed from: {exe_path.parent / 'version.yaml'}"
            )
    except Exception as e:
        if verbose:
            typer.echo(f"✗ Failed to remove SWAP executable: {e}")
        return False
    else:
        return True



import pyswap as psp

metadata = psp.components.Metadata(
    author="{author}",
    institution="{institution}",
    email="{email}",
    project="{project}",
    swap_ver="{swap_version}",
    comment="{comment}",
)



# mypy: disable-error-code="index"
from pyswap import testcase
from pyswap.db import HDF5


def test_database():
    f = HDF5(filename="data.h5")

    # Create instances of the model and result
    model = testcase.get("hupselbrook")
    model2 = model.model_copy(
        update={"version": "v2", "crop": model.crop.model_copy(update={"rds": 195})}
    )

    result = model.run("./")
    result2 = model.run("./")

    # Save to HDF5
    f.save_model(
        model=model, result=result, overwrite_datasets=True, overwrite_project=True
    )
    f.save_model(model=model2, result=result2, overwrite_datasets=True)

    # Load from HDF5
    assert f.list_projects == ["psp test - hupselbrook"]
    assert f.list_models["psp test - hupselbrook"] == ["base", "v2"]



"""Interact with YAML files.

Functions:
    load_yaml: Load a YAML file.
"""

from pathlib import Path

import yaml


def load_yaml(file: Path) -> dict:
    """Load a YAML file.

    Arguments:
        file: Path to the YAML file.
    """
    with open(file) as file:
        content: dict = yaml.safe_load(file)

    return content
